import abc
import logging

from typing import List, Optional

logger = logging.getLogger(__name__)

class Procedure(abc.ABC):
    """
    Basic unit of work to be executed by the Executor.
    It may return a list of other Procedures
    """

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    @abc.abstractmethod
    def execute(self) -> List['Procedure']:
        pass

    @abc.abstractmethod
    def rollback(self) -> None:
        pass

    @property
    def fail_plan(self) -> bool:
        if hasattr(self, "_fail_plan"):
            return self._fail_plan
        return False
    
    @fail_plan.setter
    def fail_plan(self, fail: bool) -> None:
        self._fail_plan = fail

    @property
    def name(self) -> str:
        if hasattr(self, "_name"):
            return self._name
        return "Unknown procedure"

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def friendly_name(self) -> Optional[str]:
        if hasattr(self, "_friendly_name"):
            return self._friendly_name
        return None

    @friendly_name.setter
    def friendly_name(self, friendly_name: str) -> None:
        self._friendly_name = friendly_name

    @property
    def report_callback(self) -> Optional['Procedure']:
        if hasattr(self, "_report_callback"):
            return self._report_callback
        return None

    @report_callback.setter
    def report_callback(self, report_callback: 'Procedure') -> None:
        self._report_callback = report_callback
    
    def is_async(self) -> bool:
        return False