import logging
import configparser
from os import name
import pathlib

from typing import Sequence, List, Optional

from ec2init import Plan, Procedure
from ec2init.plans.init.glassfish_das_node import ConfigureDASNode
from ec2init.plans.init.revoke_certificates import RevokeCertificates
from ec2init.plans.multinode.upgrade_oauth_config import UpgradeOAuthConfig
from ec2init.plans.plan_utils import HopsPlanUtils, REPORT_ENDPOINT, PlanUtils
from ec2init.plans.postinit.report_all_started import ReportAllStarted
from ec2init.plans.preinit.set_custom_hostname import SetCustomHostname
from ec2init.utils.bucket_configuration import BucketConfiguration
from ec2init.utils.cloud import get_cloud_provider

from ec2init.plans.init.callback_hopsworks import CallbackHopsworks
from ec2init.plans.init.issue_certificate import IssueCertificate
from ec2init.plans.init.configure_glassfish import ConfigureGlassfish
from ec2init.plans.init.start_enable_consul import StartConsul
from ec2init.plans.init.setup_updater import SetupUpdater
from ec2init.plans.init.start_services import StartServices
from ec2init.plans.init.generate_hops_certs import GenerateHopsworksEndpoint
from ec2init.plans.init.setup_eks import SetupEKS, SetupAKS, SetupGKE
from ec2init.plans.init.setup_ecr_docker_images import SetupECRDockerImages, SetupAzureDockerImages, SetupGCRDockerImages
from ec2init.plans.init.onlinefs_plan import GenerateOnlinefsApiKey
from ec2init.plans.preinit.setup_cloudwatch_agent import SetupCloudWatchAgent
from ec2init.plans.multinode.configure_oauth import ConfigureOAuth
from ec2init.plans.preinit.configure_dns import CentosConfigureDNS
from ec2init.plans.init.user_init_script import UserInitScript
from ec2init.plans.preinit.grow_data_volume import GrowDataVolume
from ec2init.utils.database import databaseUtils
from ec2init.utils.offline_backup_information import OffLineBackupInformation
from ec2init.plans.multinode.generate_elastic_signing_key import GenerateElasticSigningKey
from ec2init.utils.services import Service, ServiceUtils
from ec2init.utils.tags import TagsUtils
from ec2init.utils.docker import DockerUtils
from ec2init.utils.cloud import Provider
from ec2init.plans.init.fix_hdfs_site_for_gcp import FixHhdfSiteForGCP
from ec2init.plans.init.fix_hdfs_site_for_Azure import FixHdfsSiteForAzure
from ec2init.plans.multinode.create_service_api_key import HeadCreateServiceApiKey, WorkerCreateServiceApiKey

logger = logging.getLogger(__name__)


class BaseInitPlan(Plan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        self.name = name
        config['general']['issue_lets_encrypt'] = kwargs.get("issue_lets_encrypt", "true")
        self.config = config
        self.gateway = kwargs.get("gateway", "gateway")
        self.uid = kwargs.get("uid", "uid")
        self.instance_id = kwargs.get("instance_id", "instance_id")
        self.api_key = kwargs.get("api_key", "")
        self.user_email = kwargs.get("user_email", "user_email")
        self.user_password: Optional[str] = None
        self.password_file = pathlib.Path(kwargs.get(
            "password_file", "/root/ec2init/password"))
        if self.password_file.is_file():
            with open(self.password_file) as pwdfile:
                self.user_password = pwdfile.read().rstrip('\n')
        self.user_mgm_type = kwargs.get("user_management_type", "")
        # Default is true if not set. create_pwd_user will only be set for OAuth
        self.create_pwd_user = False if kwargs.get("create_pwd_user", "").lower() == "false" else True
        self.cloud_provider = kwargs.get("cloud_provider", "")
        self.node_id = get_cloud_provider(
            self.cloud_provider).get_node_id(config)
        kwargs['node_id'] = self.node_id
        self.plan_utils = HopsPlanUtils(config, self.user_password, **kwargs)
        self.bucket_configuration = BucketConfiguration.get_bucket_configuration(self.cloud_provider, **kwargs)
        self.region = kwargs.get("region", "")
        self.zone = kwargs.get("zone", "")
        self.gcp_service_account = kwargs.get("gcp_service_account", "")

        self.k8s_cluster_name = kwargs.get("k8s_cluster_name", "")
        self.user_registry_url = kwargs.get("user_registry_url", "")
        self.main_registry_url = kwargs.get("main_registry_url", "")
        self.docker_repos = DockerUtils.getDockerImagesInfo(kwargs.get("docker_repos", ""))
        self.python_image_name = self.config.get('initplan', 'python_image_name', fallback='python37')
        self.default_registry_url = self.config.get('initplan', 'default_registry_url', fallback='registry.service.consul:4443')
        self.cluster_domain_prefix = kwargs.get("cluster_domain_prefix", "")
        self.source_version = kwargs.get("source_version", "")
        self.is_upgrade = True if self.source_version else False 
        logger.info("%s - is_upgrade: %s", self.name, self.is_upgrade)
        # backup
        logger.info("is backup: %r", kwargs.get("is_backup", False))
        self.is_backup = bool(kwargs.get("is_backup", False))
        self.offline_backup_info = OffLineBackupInformation(config)
        offline_backup_info = self.plan_utils.get_procedure_check_file("backup_info")
        if(self.is_backup or self.is_upgrade) and self.plan_utils.check_file_exists(offline_backup_info):
            self.offline_backup_info.load(pathlib.Path(offline_backup_info))
        self.version = kwargs.get("version", "")
        logger.info("%s - version: %s", self.name, self.version)

        self.azure_storage_account = kwargs.get("azure_storage_account", "")
        self.azure_managed_identity_id = kwargs.get("azure_managed_identity_id", "")
        self.azure_acr_username = kwargs.get("azure_acr_username", "")
        self.azure_acr_password = kwargs.get("azure_acr_password", "")
        self.azure_subscription_id = kwargs.get("azure_subscription_id", "")
        self.azure_resource_group = kwargs.get("azure_resource_group", "")
        self.collect_logs = kwargs.get("collect_logs", False)
        # Upgrade
        self.deleted_blocks_retention_days = int(kwargs.get("deleted_blocks_retention_days", "0"))
        self.has_public_ip = True if kwargs.get("has_public_ip", "true").lower() == "true" else False
        logger.info("%s - has_public_ip: %s", self.name, self.has_public_ip)
        self.user_init_script = kwargs.get("user_init_script", "")
        logger.info("%s - user_init_script: %s", self.name, self.user_init_script)
        
        self.search_domain = kwargs.get("search_domain", "")
        if self.search_domain == "":
            self.search_domain = get_cloud_provider(self.cloud_provider).default_search_domain()

        self.services_list = ServiceUtils.getServicesList(kwargs.get("services_list", ""))
        self.tags = TagsUtils.getTags(kwargs.get("tags_map", ""))

        self.enable_flyingduck = bool(kwargs.get("enable_flyingduck", False))
        logger.info("%s - enable_flyingduck = %s", self.name, self.enable_flyingduck)
        self.nlb_domain = kwargs.get("nlb_domain", "")
        logger.info("%s - network_load_balancer_domain = %s", self.name, self.nlb_domain)
        self.gcp_project_id = kwargs.get("gcp_project_id", "")


    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        setup_cloudwatch_agent_check_file = self.plan_utils.get_procedure_check_file(SetupCloudWatchAgent.__name__)
        if SetupCloudWatchAgent.is_configured() and self.collect_logs and not self.plan_utils.check_file_exists(setup_cloudwatch_agent_check_file):
            if not CentosConfigureDNS.should_configure(self.plan_utils):
                logger.info(f"{self.name} - Adding procedure {SetupCloudWatchAgent.__name__}")
                cwa = SetupCloudWatchAgent(SetupCloudWatchAgent.__name__, setup_cloudwatch_agent_check_file)
                cwa.friendly_name = "Setting CloudWatch agent"
                cwa.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Setting CloudWatch agent", instance_type=self.get_instance_type())
                plan.append(cwa)

        self.add_set_custom_hostname(plan)
        return plan

    def _get_instance_type(self) -> Optional[str]:
        if hasattr(self, 'instance_type'):
            return getattr(self, 'instance_type', None)
        return None

    def get_instance_type(self) -> str:
        instance_type = self._get_instance_type()
        if instance_type is None:
            raise Exception("instance_type cannot be None")
        return str(instance_type)

    def report_error(self, error_message: str) -> None:
        PlanUtils.report_error(
            plan_name=self.name, 
            config=self.config, 
            gateway=self.gateway,
            user_id=self.uid,
            instance_id=self.instance_id,
            api_key=self.api_key,
            node_id=self.node_id,
            instance_type= self._get_instance_type(),
            error_message=error_message
        )

    # this kind of errors are not specific to one node, so return without 
    #n ode id for the error to be applied to the whole cluster
    def report_configuration_error(self, error_message: str) -> None:
        PlanUtils.report_error(
            plan_name=self.name, 
            config=self.config, 
            gateway=self.gateway,
            user_id=self.uid,
            instance_id=self.instance_id,
            api_key=self.api_key,
            error_message=error_message
        )

    def add_callback_instance_booted(self, plan: List[Procedure], user_email: Optional[str] = None, user_password: Optional[str] = None,
     user_mgm_type: Optional[str] = None) -> None:
        # Make a callback to Hopsworks Cloud indicating the instance has booted
        callback_hopsworks_check = self.plan_utils.get_procedure_check_file(
            "{}_booted".format(CallbackHopsworks.__name__))
        if not self.plan_utils.check_file_exists(callback_hopsworks_check):
            instance_type = self.get_instance_type()
            logger.info("%s - Adding procedure %s with state <booted>",
                        self.name, CallbackHopsworks.__name__)
            cp = CallbackHopsworks(CallbackHopsworks.__name__, self.config, self.gateway, self.instance_id, self.node_id, instance_type,
                                   "booted", self.uid, REPORT_ENDPOINT, self.has_public_ip, user_email, user_password, check_file=callback_hopsworks_check, api_key=self.api_key, 
                                   cloud_provider=self.cloud_provider, user_mgm_type=user_mgm_type)
            cp.fail_plan = True
            cp.friendly_name = "Report instance has booted"
            plan.append(cp)

    def add_grow_data_volume(self, plan: List[Procedure]) -> None:
        check_file = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        if not self.plan_utils.check_file_exists(check_file):
            logger.info(f"{self.name} - Adding procedure {GrowDataVolume.__name__}")
            grow_vol = GrowDataVolume(GrowDataVolume.__name__, self.config, self.plan_utils, self.is_backup)
            grow_vol.friendly_name = "Growing data volume"
            grow_vol.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Growing data volume",
                instance_type=self.get_instance_type())
            plan.append(grow_vol)

    def add_start_consul(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s",
                        self.name, StartConsul.__name__)
            cp = StartConsul(StartConsul.__name__)
            plan.append(cp)

    def add_enable_ec2update(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_enabled_ec2update_service():
            logger.info("%s - Adding procedure %s",
                        self.name, SetupUpdater.__name__)
            instance_type = self._get_instance_type()
            if instance_type is None:
                raise Exception(
                    "Could not identify instance type. Is it master or worker?")
            instance_type = str(instance_type)
            up = SetupUpdater(SetupUpdater.__name__, self.config, self.instance_id, self.node_id, instance_type,
                              self.uid, self.gateway, self.api_key, self.cloud_provider, self.has_public_ip, k8s_enabled=bool(self.k8s_cluster_name))
            up.friendly_name = "Setting up ec2update"
            up.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Setting up instance updater", instance_type=instance_type)
            plan.append(up)

    def add_start_enable_services(self, plan: List[Procedure]) -> None:
        start_services = self.plan_utils.get_procedure_check_file(
            "{}__started".format(StartServices.__name__))
        if not self.plan_utils.check_file_exists(start_services) :
            instance_type = self._get_instance_type()
            if instance_type is None:
                raise Exception("instance_type cannot be None")
            logger.info("%s - Adding procedure %s",
                        self.name, StartServices.__name__)
            ss = StartServices(StartServices.__name__, self.gateway, self.instance_id, self.node_id,
                               instance_type, self.uid, self.config, REPORT_ENDPOINT, self.api_key, self.cloud_provider, self.search_domain, check_file=start_services)
            ss.friendly_name = "Enable and start services"
            ss.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Starting services", instance_type=instance_type)
            plan.append(ss)

    def add_report_all_started(self, plan: List[Procedure]) -> None:
        instance_type = self.get_instance_type()
        report_all_started = self.plan_utils.get_procedure_check_file(ReportAllStarted.__name__)
        if not self.plan_utils.check_file_exists(report_all_started):
            logger.info(f"{self.name} - Adding procedure {ReportAllStarted.__name__} to execution plan")
            rs = ReportAllStarted(ReportAllStarted.__name__, self.config, self.plan_utils,\
                report_all_started, self.gateway, self.uid, self.instance_id, self.api_key, instance_type)
            rs.friendly_name = "Report all services have started"
            plan.append(rs)

    def _has_check_file(self, check_files: List[pathlib.Path]) -> bool:
        for file in check_files:
            if self.plan_utils.check_file_exists(file):
                return True
        return False

    def add_setup_ecr_docker_images(self, plan: List[Procedure]) -> None:
        self.add_setup_ecr_docker_images_int(plan, False)
    
    def add_setup_ecr_docker_images_int(self, plan: List[Procedure], enable_flyingduck: bool) -> None:
        if self.user_registry_url and self.main_registry_url and self.docker_repos:
            setup_docker_image_check = self.plan_utils.get_procedure_check_file(SetupECRDockerImages.__name__)
            if not self.plan_utils.check_file_exists(setup_docker_image_check) and not SetupECRDockerImages.is_already_running(SetupECRDockerImages.__name__):
                logger.info("%s - Adding procedure %s", self.name, SetupECRDockerImages.__name__)
                pd = SetupECRDockerImages(SetupECRDockerImages.__name__, check_file=setup_docker_image_check, instance_id=self.instance_id, region=self.region,
                                       ecr_user_registry_url=self.user_registry_url, ecr_main_registry_url=self.main_registry_url, ecr_repos=self.docker_repos,
                                        hopsworks_version=self.version, node_type=self.get_instance_type(), eks_cluster_enabled=bool(self.k8s_cluster_name), 
                                        offline_backup_info= self.offline_backup_info, default_registry_url=self.default_registry_url, 
                                        python_image_name=self.python_image_name, is_upgrade=self.is_upgrade, tags=self.tags, flyingduck_enabled=enable_flyingduck)
                pd.friendly_name = "Preparing Docker images for ECR"
                pd.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing Docker images for ECR")
                plan.append(pd)
        else:
            logger.info("Skipping ECR setup. No ECR Repo url specified")
    
    def add_setup_azure_docker_images(self, plan: List[Procedure]) -> None:
        if self.user_registry_url and self.main_registry_url and self.docker_repos:
            setup_docker_image_check = self.plan_utils.get_procedure_check_file(SetupAzureDockerImages.__name__)
            if not self.plan_utils.check_file_exists(setup_docker_image_check) and not SetupAzureDockerImages.is_already_running(SetupAzureDockerImages.__name__):
                logger.info("%s - Adding procedure %s", self.name, SetupAzureDockerImages.__name__)
                pd = SetupAzureDockerImages(SetupAzureDockerImages.__name__, check_file=setup_docker_image_check, instance_id=self.instance_id, region=self.region,
                                       ecr_user_registry_url=self.user_registry_url, ecr_main_registry_url=self.main_registry_url, ecr_repos=self.docker_repos,
                                        hopsworks_version=self.version, node_type=self.get_instance_type(), eks_cluster_enabled=bool(self.k8s_cluster_name), 
                                        offline_backup_info=self.offline_backup_info, default_registry_url=self.default_registry_url,
                                        python_image_name=self.python_image_name, azure_acr_username=self.azure_acr_username,
                                        azure_acr_password=self.azure_acr_password, azure_managed_identity_id=self.azure_managed_identity_id, is_upgrade=self.is_upgrade)
                pd.friendly_name = "Preparing Docker images for ACR"
                pd.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing Docker images for ACR")
                plan.append(pd)
        else:
            logger.info("Skipping ACR setup. No ACR Repo url specified")

    def add_setup_gcr_docker_images(self, plan: List[Procedure]) -> Optional[pathlib.Path]:
        if self.user_registry_url and self.main_registry_url and self.docker_repos:
            setup_docker_image_check = self.plan_utils.get_procedure_check_file(SetupGCRDockerImages.__name__)
            if not self.plan_utils.check_file_exists(setup_docker_image_check) and not SetupGCRDockerImages.is_already_running(SetupGCRDockerImages.__name__):
                logger.info("%s - Adding procedure %s", self.name, SetupGCRDockerImages.__name__)
                pd = SetupGCRDockerImages(SetupGCRDockerImages.__name__, check_file=setup_docker_image_check, instance_id=self.instance_id, region=self.region,
                                       gcr_user_registry_url=self.user_registry_url, gcr_main_registry_url=self.main_registry_url, repos=self.docker_repos,
                                        hopsworks_version=self.version, node_type=self.get_instance_type(), gke_cluster_enabled=bool(self.k8s_cluster_name), 
                                        offline_backup_info= self.offline_backup_info, default_registry_url=self.default_registry_url, 
                                        python_image_name=self.python_image_name, is_upgrade=self.is_upgrade, gcp_project_id=self.gcp_project_id)
                pd.friendly_name = "Preparing Docker images for Google Artifact Registry"
                pd.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing Docker images for Google Artifact Registry")
                plan.append(pd)
            return setup_docker_image_check
        else:
            logger.info("Skipping Google Artifact Registry setup. No Google Artifact Registry Repo url specified")
            return None

    def add_create_api_key(self, plan: List[Procedure]):
        if WorkerCreateServiceApiKey.should_run(self.plan_utils) \
            and self.plan_utils.has_started_enabled_consul():
            logger.info(f"{self.name} - Adding procedure {WorkerCreateServiceApiKey.__name__}")
            p = WorkerCreateServiceApiKey(WorkerCreateServiceApiKey.__name__, self.config, self.plan_utils, self.cloud_provider, None)
            p.friendly_name = "Creating internal API key"
            p.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Creating internal API key", instance_type=self.get_instance_type())
            plan.append(p)

    def add_generate_onlinefs_api_key(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(GenerateOnlinefsApiKey.__name__)
        if not self.plan_utils.check_file_exists(gen_check)\
            and self.plan_utils.has_configured_glassfish():
            logger.info("%s - Adding procedure %s", self.name, GenerateOnlinefsApiKey.__name__)
            gofs = GenerateOnlinefsApiKey(GenerateOnlinefsApiKey.__name__, self.config, self.api_key, gen_check)
            gofs.friendly_name = "Generating Online FS API key"
            gofs.report_callback = self.plan_utils.create_callback_procedure("starting",\
                stage="Generating online FS API key", instance_type=self.get_instance_type())
            plan.append(gofs)
        return gen_check   

    def add_centos_configure_dns(self, plan: List[Procedure]):
        if CentosConfigureDNS.should_configure(self.plan_utils):
            logger.info(f"{self.name} - Adding procedure {CentosConfigureDNS.__name__}")
            ccd = CentosConfigureDNS(CentosConfigureDNS.__name__, self.config, self.plan_utils, self.cloud_provider)
            ccd.friendly_name = "Configuring DNS"
            ccd.fail_plan = True
            plan.append(ccd)

    def add_user_init_script(self, plan: List[Procedure], call_back: bool=True) -> None:
        # run user init scrip
        user_init_check = self.plan_utils.get_procedure_check_file(UserInitScript.__name__)
        if not self.plan_utils.check_file_exists(user_init_check):
            logger.info("%s - Adding procedure %s", self.name, UserInitScript.__name__)
            userInitPlan = UserInitScript(UserInitScript.__name__, self.config, check_file=user_init_check, user_init_script=self.user_init_script)
            userInitPlan.friendly_name = "User Init Script"
            userInitPlan.fail_plan = True
            if not call_back:
                userInitPlan.report_callback = self.plan_utils.create_callback_procedure("starting", stage="User Init Script", instance_type=self.get_instance_type())
            plan.append(userInitPlan)

    def add_set_custom_hostname(self, plan: List[Procedure]):
        if SetCustomHostname.should_run(self.plan_utils, self.cloud_provider, self.search_domain):
            logger.info("%s - Adding procedure %s", self.name, SetCustomHostname.__name__)
            setCustomHostname = SetCustomHostname(SetCustomHostname.__name__, self.config, self.plan_utils, self.cloud_provider, self.search_domain)
            setCustomHostname.friendly_name = "Setting custom hostname"
            setCustomHostname.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Setting custom hostname",\
                instance_type=self.get_instance_type())
            plan.append(setCustomHostname)

    def add_fix_hdfs_site_for_bucket(self, plan: List[Procedure]):
        if self.cloud_provider == Provider.GCP and self.bucket_configuration.has_bucket():
            gcp_proc_check_file = self.plan_utils.get_procedure_check_file(FixHhdfSiteForGCP.__name__)
            if not self.plan_utils.check_file_exists(gcp_proc_check_file):
                logger.info("%s - Adding procedure %s", self.name, FixHhdfSiteForGCP.__name__)
                pro = FixHhdfSiteForGCP(FixHhdfSiteForGCP.__name__, check_file=gcp_proc_check_file,
                                   bucket_configuration=self.bucket_configuration, zone=self.zone, deleted_blocks_retention_days = self.deleted_blocks_retention_days)
                pro.friendly_name = "Fixing GCP Bucket params"
                pro.fail_plan = True
                pro.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Fixing GCP Bucket params", instance_type=self.get_instance_type())
                plan.append(pro)

        if self.cloud_provider == "azure" and self.bucket_configuration.has_bucket():
            azure_proc_check_file = self.plan_utils.get_procedure_check_file(FixHdfsSiteForAzure.__name__)
            if not self.plan_utils.check_file_exists(azure_proc_check_file):
                logger.info("%s - Adding procedure %s", self.name, FixHdfsSiteForAzure.__name__)
                p = FixHdfsSiteForAzure(FixHdfsSiteForAzure.__name__, check_file=azure_proc_check_file,
                                   container_configuration=self.bucket_configuration, storage_account=self.azure_storage_account, client_id=self.azure_managed_identity_id)
                p.friendly_name = "Fixing Azure Blob store params"
                p.fail_plan = True
                p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Fixing Azure Blob store params", instance_type=self.get_instance_type())
                plan.append(p)

class MasterBaseInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.ldap_conf_raw = kwargs.get("ldap_conf", "")
        self.ldap_conf = self.plan_utils.base64_to_json(self.ldap_conf_raw)
        self.oauth_conf_raw = kwargs.get("oauth_conf", "")
        self.oauth_conf = self.plan_utils.base64_to_json(self.oauth_conf_raw)
        self.instance_type = "master"
        self.deactivate_log_report = kwargs.get("deactivate_log_report", "").lower() == 'true'
        self.mgmd_ip = kwargs.get("ndb_mgmd_ip", "")

    def create(self) -> Sequence[Procedure]:
        plan : List[Procedure] = []
        parent_plan: Sequence[Procedure] = super().create()
        plan += list(parent_plan)
        
        self.add_grow_data_volume(plan)
        
        return plan

    # user_password argument is ignored but it has to be there to match method signature
    def add_callback_instance_booted(self, plan: List[Procedure], user_email: Optional[str] = None, user_password: Optional[str] = None,
     user_mgm_type: Optional[str] = None) -> None:
        super().add_callback_instance_booted(
            plan, user_email=self.user_email, user_password=self.user_password, user_mgm_type=self.user_mgm_type)

    def add_issue_LE_certificate(self, plan: List[Procedure]) -> None:
        if not self.plan_utils.has_issued_certificate():
            logger.info("%s - Adding procedure %s",
                        self.name, IssueCertificate.__name__)
            ilec = IssueCertificate(
                IssueCertificate.__name__, self.config, self.instance_id, self.cloud_provider, self.cluster_domain_prefix)
            ilec.friendly_name = "Issuing Let's Encrypt certificate"
            ilec.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Issuing Let's Encrypt certificate", instance_type=self.instance_type)
            plan.append(ilec)

    def add_create_api_key(self, plan: List[Procedure]):
        if HeadCreateServiceApiKey.should_run(self.plan_utils) \
            and self.plan_utils.has_configured_glassfish() \
            and self.plan_utils.has_started_enabled_consul():
            logger.info(f"{self.name} - Adding procedure {HeadCreateServiceApiKey.__name__}")
            p = HeadCreateServiceApiKey(HeadCreateServiceApiKey.__name__, self.config, self.plan_utils, self.cloud_provider)
            p.friendly_name = "Creating internal API key"
            p.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Creating internal API key", instance_type=self.instance_type)
            plan.append(p)

    def add_configure_glassfish(self, plan: List[Procedure]) -> None:
        gen_check = self.plan_utils.get_procedure_check_file(ConfigureGlassfish.__name__)
        has_configured_glassfish = self.plan_utils.has_configured_glassfish()
        if not self.plan_utils.check_file_exists(gen_check):
            logger.info("%s - Adding procedure %s", self.name, ConfigureGlassfish.__name__)
            cg = ConfigureGlassfish(ConfigureGlassfish.__name__, self.config, self.ldap_conf, gen_check, has_configured_glassfish)
            cg.friendly_name = "Configuring application server"
            cg.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Configuring application server", instance_type=self.instance_type)
            plan.append(cg)

    def add_configure_glassfish_ha(self, plan: List[Procedure]) -> Optional[pathlib.Path]:
        is_ha = PlanUtils.is_ha(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key)
        #if there are secondary nodes with GlassfishInstance
        if is_ha:
            cluster_status = PlanUtils.get_cluster_status(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key)
            head_node_ips = [node['ip'] for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.GlassfishInstance))]
            gen_check = self.plan_utils.get_procedure_check_file(ConfigureDASNode.__name__)
            has_configured_das = self.plan_utils.has_configured_glassfish_instance()
            if not self.plan_utils.check_file_exists(gen_check) \
                and self.plan_utils.has_configured_glassfish():
                logger.info("%s - Adding procedure %s", self.name, ConfigureDASNode.__name__)
                cdas = ConfigureDASNode(ConfigureDASNode.__name__, self.config, gen_check, has_configured_das, self.cloud_provider, head_node_ips, self.version)
                cdas.friendly_name = "Configuring DAS node"
                cdas.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Configuring DAS node", instance_type=self.instance_type)
                plan.append(cdas)  
            return gen_check
        return None   

    def add_revoke_old_x509_certificates(self, plan: List[Procedure], subjects: List[str]) -> None:
        check_file = self.plan_utils.get_procedure_check_file(RevokeCertificates.__name__)
        if not self.plan_utils.check_file_exists(check_file) \
            and self.plan_utils.has_configured_glassfish():
            logger.info("%s - Adding procecure %s", self.name, RevokeCertificates.__name__)
            task = RevokeCertificates(RevokeCertificates.__name__, self.config, check_file, subjects)
            task.friendly_name = "Revoking previous x.509 certificates"
            plan.append(task)
        
    def add_generate_hopsworks_internal_x509(self, plan: List[Procedure]) -> None:
        gen_check = self.plan_utils.get_procedure_check_file(
            GenerateHopsworksEndpoint.__name__)
        revoke_check = self.plan_utils.get_procedure_check_file(RevokeCertificates.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
                and self.plan_utils.has_configured_glassfish() \
                and self.plan_utils.check_file_exists(revoke_check):
            logger.info("%s - Adding procedure %s", self.name,
                        GenerateHopsworksEndpoint.__name__)
            gen_task = GenerateHopsworksEndpoint(
                GenerateHopsworksEndpoint.__name__, self.config, gen_check, self.cloud_provider, self.search_domain, self.node_id)
            gen_task.friendly_name = "Generating Hopsworks internal endpoint X.509"
            plan.append(gen_task)

    def add_setup_eks(self, plan: List[Procedure], is_primary_head : bool = True) -> None:
        if self.k8s_cluster_name:
            setup_eks_check = self.plan_utils.get_procedure_check_file(SetupEKS.__name__)
            if not self.plan_utils.check_file_exists(setup_eks_check):
                logger.info("%s - Adding procedure %s", self.name, SetupEKS.__name__)
                pks = SetupEKS(SetupEKS.__name__, check_file=setup_eks_check, region=self.region, eks_cluster_name=self.k8s_cluster_name,
                 config=self.config, is_primary_head=is_primary_head)
                pks.friendly_name = "Preparing configurations for EKS"
                pks.fail_plan = True
                pks.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing configurations for EKS")
                plan.append(pks)

    def add_setup_aks(self, plan: List[Procedure]) -> None:
        if self.k8s_cluster_name:
            setup_aks_check = self.plan_utils.get_procedure_check_file(SetupAKS.__name__)
            if not self.plan_utils.check_file_exists(setup_aks_check):
                logger.info("%s - Adding procedure %s", self.name, SetupAKS.__name__)
                pks = SetupAKS(SetupAKS.__name__, check_file=setup_aks_check, region=self.region, eks_cluster_name=self.k8s_cluster_name,
                 config=self.config, azure_resource_group=self.azure_resource_group,
                 azure_managed_identity_id=self.azure_managed_identity_id)
                pks.friendly_name = "Preparing configurations for AKS"
                pks.fail_plan = True
                pks.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing configurations for AKS")
                plan.append(pks)

    def add_setup_gke(self, plan: List[Procedure]) -> Optional[pathlib.Path]:
        if self.k8s_cluster_name:
            setup_gks_check = self.plan_utils.get_procedure_check_file(SetupGKE.__name__)
            if not self.plan_utils.check_file_exists(setup_gks_check):
                logger.info("%s - Adding procedure %s", self.name, SetupGKE.__name__)
                pks = SetupGKE(SetupGKE.__name__, check_file=setup_gks_check, region=self.region, gke_cluster_name=self.k8s_cluster_name,
                 config=self.config, gcp_project_id=self.gcp_project_id)
                pks.friendly_name = "Preparing configurations for GKE"
                pks.fail_plan = True
                pks.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing configurations for GKE")
                plan.append(pks)
            return setup_gks_check
        return None

    def add_configure_oauth(self, plan: List[Procedure]) -> Optional[pathlib.Path]:
        if self.oauth_conf:
            if not self.is_upgrade and not self.is_backup:
                gen_check = self.plan_utils.get_procedure_check_file(ConfigureOAuth.__name__)
                if not self.plan_utils.check_file_exists(gen_check) \
                    and self.plan_utils.has_configured_glassfish() \
                    and self.plan_utils.has_started_enabled_consul():
                    logger.info("%s - Adding procedure %s", self.name, ConfigureOAuth.__name__)
                
                    oauth = ConfigureOAuth(ConfigureOAuth.__name__, self.config, gen_check, self.oauth_conf, self.has_public_ip,  databaseUtils.ADMIN_EMAIL,
                    self.user_password, self.instance_id, self.cloud_provider, self.cluster_domain_prefix)
                    oauth.friendly_name = "Configuring OAuth client"
                    oauth.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Configuring OAuth client", instance_type=self.instance_type)
                    plan.append(oauth)
                return gen_check
            elif self.is_upgrade:
                check_file = self.plan_utils.get_procedure_check_file(UpgradeOAuthConfig.__name__)
                if not self.plan_utils.check_file_exists(check_file) \
                    and self.plan_utils.has_configured_glassfish() \
                    and self.plan_utils.has_started_enabled_consul():
                    logger.info("%s - Adding procedure %s", self.name, UpgradeOAuthConfig.__name__)
                
                    upgradeOauth = UpgradeOAuthConfig(UpgradeOAuthConfig.__name__, check_file, self.oauth_conf, self.user_mgm_type)
                    upgradeOauth.friendly_name = "Upgrade OAuth client configuration"
                    upgradeOauth.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Upgrade OAuth client configuration", instance_type=self.instance_type)
                    plan.append(upgradeOauth)
                return check_file
        return None


    def add_generate_elastic_signing_key(self, plan: List[Procedure]) -> pathlib.Path:
        gen_check = self.plan_utils.get_procedure_check_file(
            GenerateElasticSigningKey.__name__)
        if not self.plan_utils.check_file_exists(gen_check) \
            and self.plan_utils.has_configured_glassfish() \
            and self.plan_utils.has_started_enabled_consul():
            logger.info("%s - Adding procedure %s", self.name, GenerateElasticSigningKey.__name__)
            jwtp = GenerateElasticSigningKey(GenerateElasticSigningKey.__name__, self.config, gen_check)
            jwtp.friendly_name = "Generating elastic signing key"
            jwtp.report_callback = self.plan_utils.create_callback_procedure(
                "starting", stage="Generating elastic signing key", instance_type=self.instance_type)
            plan.append(jwtp)
        return gen_check
