import logging
import typing
import boto3
import filecmp


from typing import Dict, List
from mypy_boto3_s3 import Client

from ec2init import Procedure
from ec2init.utils.bucket_configuration import BucketConfiguration, S3BucketEncryptionConfiguration, S3GeneralConfiguration
from pathlib import Path
from ec2init.utils.exceptions import ConfigException

logger = logging.getLogger(__name__)

class CheckS3Permissions(Procedure):
    def __init__(self, name: str, check_file: Path, bucket_configuration: BucketConfiguration, region: str, s3_test_obj: Path, is_backup_or_upgrade: bool,
                deleted_blocks_retention_days: int) -> None:
        self.name = name
        self.check_file = check_file
        self.bucket_configuration = bucket_configuration
        self.s3_test_obj = s3_test_obj
        self.region = region
        self.is_backup_or_upgrade = is_backup_or_upgrade
        self.deleted_blocks_retention_days = deleted_blocks_retention_days

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def constructExtraArgs(self) -> Dict[str, str]:
        extraArgs: Dict[str, str] = {}
        if self.bucket_configuration.encryption_configuration.is_encryption_enabled():
            encryptionProperties = typing.cast(S3BucketEncryptionConfiguration, self.bucket_configuration.encryption_configuration)
            if encryptionProperties.is_kms_enabled():
                extraArgs['ServerSideEncryption'] = "aws:kms"
                # boto3 1.14.18 does not understand this property
                # extraArgs['BucketKeyEnabled'] = str(encryptionProperties.bucket_key)
                if not encryptionProperties.is_managed_kms():
                    extraArgs['SSEKMSKeyId'] = encryptionProperties.user_key_arn
            else:
                extraArgs['ServerSideEncryption'] = "AES256"

        if self.bucket_configuration.general_configuration:
            s3_general_conf = typing.cast(S3GeneralConfiguration, self.bucket_configuration.general_configuration)
            if s3_general_conf.acl_bucket_full_control:
                extraArgs['ACL'] = "bucket-owner-full-control"

        return extraArgs

    def execute(self) -> List[Procedure]:
        logger.info("Checking if this instance has proper permissions to access S3 . Retention days "+ str(self.deleted_blocks_retention_days))
        bucket_name = self.bucket_configuration.name
        try:
            s3: Client = boto3.client('s3')

        # check that the bucket and VMs are in the same region            
        #s3.get_bucket_location only works if one is the owner of the bucket (https://docs.aws.amazon.com/AmazonS3/latest/API/API_GetBucketLocation.html) 
        #using a workaround suggested here https://github.com/aws/aws-sdk-cpp/issues/844
            info = s3.head_bucket(Bucket=bucket_name) # type: ignore
        except Exception as ex:
            logger.error("Error while checking connectivity to the bucket", exc_info=ex)
            raise ConfigException(f"Unable to access to the S3 bucket ({bucket_name}). Make sure that the bucket name is correct and that you have added the " +
                "S3:ListBucket permission to the instance profile. " +
                "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile") 

        self.bucket_location = str((info['ResponseMetadata']['HTTPHeaders']['x-amz-bucket-region']))
        if self.bucket_location != self.region:
            raise ConfigException("The bucket you provided is not in the same region as the cluster. Make sure to run the cluster in the same region as the cluster. " + 
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-2-creating-storage") 

        # HopsFS 3.2.0.3 now requires checking support for versioning even if backups are not enabled. 
        self.isVersioningEnabled(s3)

        if self.deleted_blocks_retention_days > 0:
            if not self.isVersioningEnabled(s3) :
                self.enableVersioning(s3)

            #put block retention policy
            self.putBlockRetentionPolicy(s3)

        #Check listing permission 
        try:
            resp = s3.list_objects_v2(Bucket=bucket_name)
        except Exception as ex:
            logger.error("Error listing the objects in the bucket", exc_info=ex)
            raise ConfigException("Unable to list the bucket content. " +
            "Make sure that you have added the S3:ListBucket permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")
        
        #Check empty bucket
        keys = resp['KeyCount'] > 0
        if keys > 0 and not self.is_backup_or_upgrade:
            raise ConfigException("The bucket you provided is not empty. Please empty the bucket and retry or create a new cluster with a different bucket.")

        #Check write permission
        try:
            key = "test-obj"
            with open(self.s3_test_obj, 'rb') as data:
                s3.upload_fileobj(data, bucket_name, key, ExtraArgs=self.constructExtraArgs())
        except Exception as ex:
            logger.error("Error puting an object in the bucket", exc_info=ex)
            raise ConfigException("Unable to put an object in the bucket. " +
            "Make sure that you have added the S3:PutObject permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

        #Check read permission
        try:
            newFile = str(self.s3_test_obj) + ".new" 
            s3.download_file(bucket_name, key, newFile)
        except Exception as ex:
            logger.error("Error getting an object from the bucket", exc_info=ex)
            raise ConfigException("Unable to get an object from the bucket. " +
            "Make sure that you have added the S3:GetObject permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

        if not (filecmp.cmp(str(self.s3_test_obj), newFile)):
            logger.error("Error getting an object from the bucket, the object got corrupted")
            raise ConfigException("Unable to get an object from the bucket. The object got corrupted " +
            "Make sure that nothing else is accessing the bucket and retry. If the error persist contact us.")

        try:
            s3.delete_object(Bucket=bucket_name, Key=key)
        except Exception as ex:
            logger.error("Error deleting an object from the bucket", exc_info=ex)
            raise ConfigException("Unable to delete an object from the bucket. " +
            "Make sure that you have added the S3:DeleteObject permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

        if self.deleted_blocks_retention_days > 0:
            try:
                versions_list = s3.list_object_versions(Bucket=bucket_name, Prefix=key)
            except Exception as ex:
                logger.error("Error listing an object version", exc_info=ex)
                raise ConfigException("Unable to list the versions of an object in the bucket. " +
                "Make sure that you have added the S3:ListBucketVersions permission to the instance profile. " +
                "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

            try:
                latest_version = versions_list['DeleteMarkers'][0]['VersionId']
                s3.delete_object(Bucket=bucket_name, Key=key, VersionId=latest_version)
            except Exception as ex:
                logger.error("Error listing an object version", exc_info=ex)
                raise ConfigException("Unable to delete the versions of an object in the bucket. " +
                "Make sure that you have added the S3:DeleteObjectVersion permission to the instance profile. " +
                "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")
        
            #deleting the object version put back the object in the bucket, clean that up
            s3.delete_object(Bucket=bucket_name, Key=key)

        try:
            multipart_obj_key='multipart-upload-test'
            multipart_args: typing.Mapping[str, str] = self.constructExtraArgs()
            resp_cmu = s3.create_multipart_upload(Bucket=bucket_name, Key=multipart_obj_key, **multipart_args)
        except Exception as ex:
            logger.error("Error creating a multipart upload to the bucket", exc_info=ex)
            raise ConfigException("Unable to creat a multipart upload to the bucket. " +
            "Make sure that you have added all the permissions to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

        try:
            resp_up = s3.upload_part(
                Body=b'some bytes',
                Bucket=bucket_name,
                Key=multipart_obj_key,
                PartNumber=1,
                UploadId=resp_cmu['UploadId']
            )
        except Exception as ex:
            logger.error("Error uploading a part to the bucket", exc_info=ex)
            raise ConfigException("Unable to upload a multipart object to the bucket. " +
            "Make sure that you have added all the permissions to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")


        try:
            resp_lp = s3.list_multipart_uploads(
                Bucket=bucket_name
            )
        except Exception as ex:
            logger.error("Error listing the multipart uploads on the bucket", exc_info=ex)
            raise ConfigException("Unable list the multipart upload for the bucket. " +
            "Make sure that you have added the S3:ListBucketMultipartUploads permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

        
        s3.complete_multipart_upload(
            Bucket=bucket_name,
            Key=multipart_obj_key,
            MultipartUpload={
                'Parts': [
                    {
                        'ETag': resp_up['ETag'],
                        'PartNumber': 1 
                    },
                ]
            },
            UploadId=resp_cmu['UploadId']
        )

        s3.delete_object(Bucket=bucket_name, Key=multipart_obj_key)
        
        
        resp_cmu = s3.create_multipart_upload(Bucket=bucket_name, Key=multipart_obj_key, **multipart_args)

        try:
            s3.abort_multipart_upload(
                Bucket=bucket_name,
                Key=multipart_obj_key,
                UploadId=resp_cmu['UploadId']
            )
        except Exception as ex:
            logger.error("Error aborting the multipart uploads on the bucket", exc_info=ex)
            raise ConfigException("Unable abort a multipart upload on the bucket. " +
            "Make sure that you have added the S3:AbortMultipartUpload permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")
        logger.info("All S3 Permission checks PASSED")

        #Every thing is OK. 
        if self.check_file is not None:
            self.check_file.touch()

        return []


    def rollback(self) -> None:
        pass

    def putBlockRetentionPolicy(self, s3: Client) -> None:
        try:
            response = s3.put_bucket_lifecycle_configuration( # type: ignore 
                Bucket=self.bucket_configuration.name,
                LifecycleConfiguration={
                'Rules': [
                    {
                    'Expiration': {
                        'ExpiredObjectDeleteMarker': True
                    }, 
                    'ID': 'HopsFSBlocksRetentionPolicy',
                    'Filter': {},
                    'Status': 'Enabled',
                    'NoncurrentVersionExpiration': {
                        'NoncurrentDays': self.deleted_blocks_retention_days 
                    }, 
                    },
                ],
                },
            ) 
            
            logger.info("Put block retention policy response "+str(response))
            if response['ResponseMetadata']['HTTPStatusCode'] != 200:
                raise Exception("Unable to set block retention policy to the S3 bucket. Response: "+str(response))
        except Exception as ex:
            logger.error("Error putting block retention policy", exc_info=ex)
            raise ConfigException("Unable to set the block retention policy. " +
            "Make sure that you have added S3:PutLifecycleConfiguration permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")


    def enableVersioning(self, s3: Client) -> None:
        try:
            response = s3.put_bucket_versioning( # type: ignore 
                Bucket=self.bucket_configuration.name,
                VersioningConfiguration={
                    'Status': 'Enabled',
                },
            )
            logger.info("Enable versioning response "+str(response))
            if response['ResponseMetadata']['HTTPStatusCode'] != 200:
                raise Exception("Unable to set  bucket versioning. Response: "+str(response))
        except Exception as ex:
            logger.error("Error enabling versioning on the bucket", exc_info=ex)
            raise ConfigException("Unable to enable the bucket versioning. " +
            "Make sure that you have added S3:PutBucketVersioning permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile")

    def isVersioningEnabled(self, s3: Client) -> bool:
        try:
            response = s3.get_bucket_versioning( 
                Bucket=self.bucket_configuration.name,
            )
            logger.info("Get versioning response "+str(response))
            if response['ResponseMetadata']['HTTPStatusCode'] != 200:  # type: ignore
                raise Exception("Unable to get  bucket versioning. Response: "+str(response))
            else: 
                if 'Status' in response and response['Status']  == 'Enabled':
                    return True
                else:
                    return False
        except Exception as ex:
            logger.error("Error while getting the bucket version", exc_info=ex)
            raise ConfigException("Unable to get the bucket versioning information. Make sure that you have added the S3:GetBucketVersioning permission to the instance profile. " +
            "See our documentation for more details: https://docs.hopsworks.ai/latest/setup_installation/aws/getting_started/#step-3-creating-instance-profile") 
        
            
        