import logging
import tempfile
import shutil
import os
import json
import ec2init.utils.cryptography as utils_crypto
from cryptography.x509.oid import NameOID

from urllib.parse import urlparse
from configparser import ConfigParser
from pathlib import Path

from typing import List, Callable, Tuple, Dict, cast

from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from ec2init.utils.dns import DnsUtils
from ec2init.utils.http import HttpUtils
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.constants import Url
from ec2init.utils.common import CommonUtils

logger = logging.getLogger(__name__)


class GenerateHopsCerts(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, cloud_provider: str, search_domain: str, node_id: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.cloud_provider = cloud_provider
        self.search_domain = search_domain
        self.node_id = node_id

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning(
                    "%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)

        # Get my FQDN
        fqdn = get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)
        logger.debug("%s - My FQDN is: %s", self.name, fqdn)
        kagent_conf_dir = self.config.get(
            'general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf = Path(kagent_conf_dir, "config.ini")
        _, path = tempfile.mkstemp(text=True)
        try:
            shutil.copy(kagent_conf, path)
            tmp_config = ConfigParser()
            tmp_config.read(str(path))

            self._wait_for_hopsworks_domain(tmp_config)

            certs_user = tmp_config.get(
                'agent', 'certs-user', fallback='certs')
            tmp_config.set('agent', 'hostname', fqdn)
            with open(str(path), 'w') as fd:
                tmp_config.write(fd)
            kagent_config_group = kagent_conf.group()
            shutil.chown(path, group=kagent_config_group)
            os.chmod(path, 0o750)
            logger.debug(
                "%s - Written temporary configuration file", self.name)
            logger.debug(
                "%s - Starting X.509 re-generation with CN=%s", self.name, fqdn)

            # Load Crypto state store and for each user rotate the certificates
            def cn_gen(cn): return fqdn if cn == "localhost" or cn == "glassfish.service.consul" else cn
            self._call_hopsify(path, tmp_config, fqdn, cn_gen)
        finally:
            os.remove(path)

        # Restart services to load new X.509 certificates
        self._restart_required_services()

        if self.check_file is not None:
            self.check_file.touch()
        return []

    def _wait_for_hopsworks_domain(self, kagent_config: ConfigParser) -> None:
        # Wait until we can resolve Hopsworks domain name
        raw_hw_url = kagent_config.get(
            'server', 'url', fallback='hopsworks.glassfish.service.consul')
        hw_url = urlparse(raw_hw_url)
        DnsUtils.get_ip_from_name(str(hw_url.hostname))
        logger.info("%s - Resolved Hopsworks domain: %s",
                    self.name, hw_url.hostname)

    def _restart_required_services(self) -> None:
        self._restart_mysqld()
        self._restart_consul()
        self._restart_docker_registry()

    def _restart_mysqld(self) -> None:
        logger.info("%s - Restarting mysqld", self.name)
        command = ["systemctl", "restart", "mysqld"]
        SystemUtils.execute_command(command)
        logger.info("%s - Restarted mysqld", self.name)

    def _restart_consul(self) -> None:
        logger.info("%s - Restarting consul", self.name)
        command = ["systemctl", "restart", "consul"]
        SystemUtils.execute_command(command)
        logger.info("%s - Restarted consul", self.name)

    def _restart_docker_registry(self) -> None:
        logger.info("%s - Restarting docker registry if exists", self.name)
        command = ["docker", "restart", "registry"]
        try:
            SystemUtils.execute_command(command)
        except Exception as ex:
            logger.info("%s - Failed to restart docker registry : %s", self.name, ex)
        
    def _call_hopsify(self, agentConfigPath: str, agentConfig: ConfigParser, fqdn: str,
                      cn_gen: Callable[[str], str]) -> None:
        certs_user = agentConfig.get('agent', 'certs-user', fallback='certs')
        run_csr_script = Path(agentConfig.get(
            'agent', 'sbin-dir', fallback='/srv/hops/sbin'), "run_csr.sh")
        self.fix_elkadmin_OU(agentConfig)
        users_cn = self.get_users_common_name(agentConfig)
        for user, cn in users_cn:
            # Special case when turning from localhost installation
            effective_cn = cn_gen(cn)
            logger.info("%s - Rotating X.509 for user %s", self.name, user)

            gen_x509_cmd_generator = self._get_gen_x509_command_generator()
            gen_x509_cmd = gen_x509_cmd_generator(certs_user, str(
                run_csr_script), agentConfigPath, user, effective_cn)
            logger.debug("%s - Running command %s",
                         self.name, " ".join(gen_x509_cmd))

            SystemUtils.execute_command(gen_x509_cmd, timeout=900)
            logger.info(
                "%s - Done re-generating certificate with CN=%s for user %s", self.name, fqdn, user)

    def _get_gen_x509_command_generator(self) -> Callable[[str, str, str, str, str], List[str]]:
        return lambda certs_user, run_script, agent_config_file, user, cn: ["sudo", "-u", certs_user, run_script, "--config", agent_config_file,
                                                                            "x509", "--rotation", "--username", user, "--common-name", cn, "--skip-revoke"]

    def get_users_common_name(self, agentConfig: ConfigParser) -> List[Tuple[str, str]]:
        users_cn: List[Tuple[str, str]] = []
        state_store_location = agentConfig.get(
            'agent', 'state-store', fallback='/srv/hops/kagent/etc/state_store')
        state_store = os.path.join(
            state_store_location, "crypto_material_state.json")
        with open(state_store, "r") as fd:
            hopsify_state = json.load(fd)
        crypto_state = hopsify_state["crypto_state_store"]
        users = [u.strip() for u in crypto_state]
        
        for user in users:
            cn = crypto_state[user]['common_name']
            users_cn.append((user, cn))
        return users_cn

    def fix_elkadmin_OU(self, agentConfig: ConfigParser):
        state_store_location = agentConfig.get(
            'agent', 'state-store', fallback='/srv/hops/kagent/etc/state_store')
        state_store = os.path.join(
            state_store_location, "crypto_material_state.json")
        if os.path.isfile(state_store):
            with open(state_store, "r") as fd:
                hopsify_state = json.load(fd)
            if "elkadmin" in hopsify_state["crypto_state_store"]:
                hopsify_state["crypto_state_store"]["elkadmin"]["crypto_version"] = 0
                with open(state_store, 'w') as fd:
                    json.dump(hopsify_state, fd)

    def rollback(self) -> None:
        return

class GenerateHopsworksEndpoint(Procedure):
    """
    Task to regenerate Hopsworks internal endpoint certificate
    """

    def __init__(self, name: str, config: ConfigParser, check_file: Path, cloud_provider: str, search_domain: str, node_id: str) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.cloud_provider = cloud_provider
        self.search_domain = search_domain
        self.node_id = node_id

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def execute(self) -> List[Procedure]:
        logger.info("%s - Generating Hopsworks internal endpoint X.509 certificate")

        logger.debug("%s - Generating private key", self.name)
        private_key = utils_crypto.generate_private_key(2048)
        serialized_private_key = utils_crypto.serialize_private_key_rsa(private_key)
        glassfish_config = "/srv/hops/domains/domain1/config"

        pkey_file = Path(glassfish_config, "internal.key")
        with open(pkey_file, 'wb') as fd:
            fd.write(serialized_private_key)
        logger.info("%s - Finished generating Hopsworks internal endpoint X.509", self.name)

        logger.debug("%s - Generating CSR", self.name)
        #Python 3, all strings are unicode
        fqdn = get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)
        csr = utils_crypto.generate_certificate_signing_request(
            private_key,
            {
                NameOID.COMMON_NAME: fqdn,
                NameOID.LOCALITY_NAME: u'glassfishinternal',
                NameOID.ORGANIZATIONAL_UNIT_NAME: u'0'
            }
        )
        serialized_csr = utils_crypto.serialize_certificate_signing_request(csr)
        logger.debug("%s - Generated CSR", self.name)

        agent_username, agent_password = CommonUtils._get_agent_credentials(self.config)
        if not agent_username or not agent_password:
            raise Exception("Agent username or password is not set")

        method, uri = Url.HOPSWORKS_LOGIN
        uri = urlparse(uri)._replace(netloc="127.0.0.1").geturl()
        data = {
            'email': agent_username,
            'password': agent_password
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        logger.debug("%s - Logging in to Hopsworks", self.name)

        response: Dict[str, str] = HttpUtils.do_plain_call(self.config, uri, method, data=data, headers=headers, verify=False)
        response_headers: Dict[str, str] = cast(Dict[str, str], response['headers'])
        token = response_headers['Authorization']

        body = {
            'csr': serialized_csr.decode("utf-8")
        }
        headers['Content-Type'] = "application/json"
        headers['Authorization'] = token
        method, uri = Url.HOST_CA_GENERATE

        logger.debug("%s - Sending CSR for signing", self.name)
        # At this point CA has not been initialized yet
        # Increase the timeout because initialization takes some time
        certificate_response = HttpUtils.do_api_call(self.config, uri, method, body, headers=headers, verify=False, timeout=120)
        logger.debug("%s - Gotten signed certificate", self.name)

        signed_certificate = certificate_response['signedCert']
        intermediateCACertificate = certificate_response['intermediateCaCert']
        rootCACertificate = certificate_response['rootCaCert']
        certificate_bundle = f"{signed_certificate}\n{intermediateCACertificate}"
        
        files = {
            Path(glassfish_config, "signed_certificate.pem"): signed_certificate,
            Path(glassfish_config, "intermediate_ca.pem"): intermediateCACertificate,
            Path(glassfish_config, "certificate_bundle.pem"): certificate_bundle,
            Path(glassfish_config, "root_ca.pem"): rootCACertificate,
        }

        for f, c in files.items():
            logger.info("%s - Writing file %s", self.name, f)
            with open(str(f), 'w') as cfd:
                cfd.write(c)

        installation_dir = self.config.get('general', 'installation_dir', fallback='/root/ec2init')
        import_internal_script = Path(installation_dir, "import_glassfish_internal_x509.sh")
        logger.info("%s - Importing Hopsworks internal endpoint X.509 to keystores", self.name)
        SystemUtils.execute_command(["/bin/bash", str(import_internal_script)], timeout=120)
        logger.info("%s - Finished importing Hopsworks internal endpoint X.509 to keystores", self.name)

        if self.check_file:
            self.check_file.touch()
            
        return []

    def rollback(self) -> None:
        pass
