import logging
import configparser
import pathlib
from urllib.parse import urlparse

from typing import List, Dict, cast

from ec2init import Procedure
from ec2init.utils.http import HttpUtils
from ec2init.utils.dns import DnsUtils
from ec2init.utils.constants import Url
from ec2init.utils.random_generator import RandomGenerator

logger = logging.getLogger(__name__)

class GenerateOnlinefsApiKey(Procedure):

    def __init__(self, name: str, config: configparser.ConfigParser, onlinefs_password: str, check_file: pathlib.Path) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file
        self.onlinefs_password = onlinefs_password
        self.rand = RandomGenerator()

    def rollback(self) -> None:
        pass
    
    def execute(self) -> List[Procedure]:
        # Login to Hopsworks to get JWT
        onlinefs_username = self.config.get('initplan', 'onlinefs_username', fallback='onlinefs@hopsworks.ai')
        login_params = {
            'email': onlinefs_username,
            'password': self.onlinefs_password
        }

        method, uri = Url.HOPSWORKS_LOGIN

        # For safety wait until Consul domain name becomes available
        # In Head node this is not an issue as the Procedure is executed
        # only after Glassfish has been configured but for other nodes it
        # can be problematic
        u = urlparse(uri)
        if u.hostname is None:
            raise Exception(f"Cannot parse hostname from {uri}")
        DnsUtils.get_ip_from_name(u.hostname)
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        response: Dict[str, str] = HttpUtils.do_plain_call(self.config, uri, method, data=login_params, headers=headers, verify=False)
        response_headers: Dict[str, str] = cast(Dict[str, str], response['headers'])
        jwt = response_headers['Authorization']

        # Make request to generate an API key
        api_key_name = f"onlinefs_{self.rand.random_string(12)}"
        api_key_params = {
            'name': api_key_name,
            'scope': 'KAFKA'
        }
        headers = {
            'Authorization': jwt
        }
        method, uri = Url.NEW_HOPSWORKS_API_KEY_URL
        api_key_response: Dict[str, str] = HttpUtils.do_api_call(self.config, uri, method, params=api_key_params, headers=headers, verify=False)
        api_key = api_key_response['key']

        # Finally write the API key
        onlinefs_token_file = self.config.get('initplan', 'onlinefs_token_file', fallback='/srv/hops/onlinefs/etc/token')
        with open(onlinefs_token_file, 'w') as fd:
            fd.write(api_key)

        self._done()
        
        return []

    def _done(self) -> None:
        if self.check_file:
            self.check_file.touch()

    