import logging
import xml.etree.ElementTree as ET
import re

from configparser import ConfigParser
from pathlib import Path

from typing import List, Dict, Any, cast, Optional

from ec2init import Procedure
from ec2init.utils.http import HttpUtils
from ec2init.utils.dns import DnsUtils
from ec2init.plans.preinit.pre_inits import BasePreInit
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.database import databaseUtils

logger = logging.getLogger(__name__)
LOGIN_URI = ("POST", "https://hopsworks.glassfish.service.consul:8182/hopsworks-api/api/auth/login")
CREATE_CLIENT_URI = ("POST", "https://hopsworks.glassfish.service.consul:8182/hopsworks-remote-user-api/api/remote/user/oauth/client")

class ConfigureOAuth(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, oauth_conf: Dict[str, Any], 
                 has_public_ip: bool, user_email: str, user_password: Optional[str], cluster_id: str, cloud_provider: str = "", cluster_domain_prefix: str = "") -> None:
        self.name = name
        self.check_file = check_file
        self.user_email = user_email
        self.user_password = user_password
        self.oauth_conf = oauth_conf
        self.has_public_ip = has_public_ip
        self.cloud_provider = cloud_provider
        self.cluster_id = cluster_id
        self.config = config
        self.cluster_domain_prefix = cluster_domain_prefix

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)

    def _login(self):
        if not self.user_email or not self.user_password:
            raise Exception("Admin username or password is not set")

        method, uri = LOGIN_URI
        data = {
            'email': self.user_email,
            'password': self.user_password,
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        response: Dict[str, str] = HttpUtils.do_plain_call(self.config, uri, method, data=data, headers=headers, verify=False)
        response_headers: Dict[str, str] = cast(Dict[str, str], response['headers'])

        jwt = response_headers['Authorization']
        return jwt

    def _create_client(self):

        method, uri = CREATE_CLIENT_URI
        jwt = self._login() #can we save the jwt for retry?

        headers = {
            'Content-Type': 'application/json',
            'Authorization': jwt
        }

        if self.has_public_ip:
            public_domain_name = get_cloud_provider(self.cloud_provider).get_public_domain_name(self.config, self.cluster_id, self.cluster_domain_prefix)
            self.oauth_conf['callbackURI'] = 'https://' + public_domain_name # is there a way to get the Scheme
        else:
            self.friendly_name = "OAuth configuration needs a public IP."
            raise Exception(self.friendly_name)

        HttpUtils.do_api_call(self.config, uri, method, data=self.oauth_conf, headers=headers, verify=False)
        

    def execute(self) -> List[Procedure]:
        if self.oauth_conf:
            logger.info("%s - Starting execution", self.name)

            DnsUtils.get_ip_from_name("hopsworks.glassfish.service.consul")

            self._create_client()

            if self.check_file is not None:
                self.check_file.touch()
            logger.info("%s - Finished execution", self.name)
        return []

    def rollback(self) -> None:
        pass
