from ec2init import Procedure

from configparser import ConfigParser
from pathlib import Path
from typing import List, Dict, cast
import logging
from ec2init.utils.constants import Url
from ec2init.utils.common import CommonUtils
from ec2init.utils.http import HttpUtils
from ec2init.utils.database import databaseUtils
import yaml 
from yaml.loader import SafeLoader

logger = logging.getLogger(__name__)

class GenerateElasticSigningKey(Procedure):
    
    GEN_ELK_TOKEN_URI = ("GET", "https://hopsworks.glassfish.service.consul:8182/hopsworks-api/api/jwt/elk/key")

    def __init__(self, name: str, config: ConfigParser, check_file: Path) -> None:
        self.name = name
        self.config = config
        self.check_file = check_file

    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)
        signing_key = self._get_elk_signing_key()
        config_yml = "/srv/hops/opensearch/plugins/opensearch-security/securityconfig/config.yml"
        with open(config_yml, 'r') as f:
            config_dict : Dict  = yaml.load(f, Loader=SafeLoader)
        config_dict['config']['dynamic']['authc']['jwt_auth_domain']['http_authenticator']['config']['signing_key'] = signing_key
        with open(config_yml, 'w') as f:
            yaml.dump(config_dict, f, sort_keys=False, default_flow_style=False)
        
        if self.check_file is not None:
            self.check_file.touch()
        logger.info("%s - Finished execution", self.name)
        return []

    def _get_elk_signing_key(self) -> str:
        databaseUtils.run_delete_statement("jwt_signing_key", "name = 'elk_jwt_signing_key'")

        master_token = self.get_service_jwt_tokens()
        headers = {
            'Content-Type': 'application/json',
            'Authorization': f"Bearer {master_token}"
        }

        method, uri = GenerateElasticSigningKey.GEN_ELK_TOKEN_URI

        response: Dict[str, str] = HttpUtils.do_plain_call(self.config, uri, method, headers=headers, verify=False)
        return response['response']

    def get_service_jwt_tokens(self) -> str:
        agent_username, agent_password = CommonUtils._get_agent_credentials(self.config)
        if not agent_username or not agent_password:
            raise Exception("Agent username or password is not set")

        method, uri = Url.HOPSWORKS_LOGIN
        data = {
            'email': agent_username,
            'password': agent_password,
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        response: Dict[str, str] = HttpUtils.do_plain_call(self.config, uri, method, data=data, headers=headers, verify=False)
        response_headers: Dict[str, str] = cast(Dict[str, str], response['headers'])
        # This is form 'Bearer TOKEN'
        master_bearer_token = response_headers['Authorization']
        master_token = master_bearer_token.split(" ")[1].strip()

        return master_token

    def rollback(self) -> None:
        pass
