import logging
import pathlib

from configparser import ConfigParser

from typing import Sequence, List
from ec2init.plans.init.revoke_certificates import RevokeCertificates

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import MasterBaseInitPlan

from ec2init.plans.multinode.master_seds import MasterSeds
from ec2init.plans.multinode.register_generate_certs import RegisterGenerateCerts
from ec2init.utils import cloud
from ec2init.plans.init.deactivate_log_report import DeactivateLogReport
from abc import abstractmethod
from ec2init.plans.init.setup_ecr_docker_images import SetupECRDockerImages, SetupAzureDockerImages, SetupGCRDockerImages
from ec2init.plans.init.restore_deleted_blocks import RestoreDeletedBlocks
from ec2init.plans.multinode.create_service_api_key import CreateServiceApiKey
from ec2init.plans.init.configure_smart_devices_manager import ConfigureSmartDeviceManagerOnGCR

logger = logging.getLogger(__name__)

class MasterInitPlan(MasterBaseInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_fix_hdfs_site_for_bucket(plan)

        # Replace master node IP placeholder with the actual IP
        sed_check = self.plan_utils.get_procedure_check_file(MasterSeds.__name__)
        if not self.plan_utils.check_file_exists(sed_check):
            logger.info("%s - Adding procedure %s", self.name, MasterSeds.__name__)
            sed_task = MasterSeds(MasterSeds.__name__, self.config, sed_check, self.cloud_provider,
                self.instance_id, self.search_domain, self.node_id, self.cluster_domain_prefix, self.gateway, self.uid, self.instance_id, self.api_key, self.deactivate_log_report)
            sed_task.friendly_name = "Updating addresses"
            sed_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Updating addresses", instance_type = self.instance_type)
            plan.append(sed_task)

        if self.deactivate_log_report:
            deactivate_log_report_check = self.plan_utils.get_procedure_check_file(DeactivateLogReport.__name__)
            if not self.plan_utils.check_file_exists(deactivate_log_report_check):
                logger.info("%s - Adding procedure %s", self.name, DeactivateLogReport.__name__)
                deactivate_log_report__task = DeactivateLogReport(DeactivateLogReport.__name__, self.config, deactivate_log_report_check)
                deactivate_log_report__task.friendly_name = "Deactivate log report"
                deactivate_log_report__task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Deactivating log report", instance_type = self.instance_type)
                plan.append(deactivate_log_report__task)

        # In case of backup restore we would like to recover S3 bucket
        # In case of upgrade, we take a backup of the data disk and right after we do the restore - we don't enforce the users to enable versioing on their bucket enabled
        if self.is_backup and not self.is_upgrade:
            restore_deleted_blocks_check = self.plan_utils.get_procedure_check_file(RestoreDeletedBlocks.__name__)
            if not self.plan_utils.check_file_exists(restore_deleted_blocks_check):
                logger.info("%s - Adding procedure %s", self.name, RestoreDeletedBlocks.__name__)
                restore_deleted_blocks_task = RestoreDeletedBlocks(RestoreDeletedBlocks.__name__, self.config, restore_deleted_blocks_check)
                restore_deleted_blocks_task.friendly_name = "Restoring deleted blocks"
                restore_deleted_blocks_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Restoring delete blocks", instance_type = self.instance_type)
                plan.append(restore_deleted_blocks_task)
            

        # Issue Let's Encrypt certificate and setup
        self.add_issue_LE_certificate(plan)

        # Configure Glassfish to use previously issued certificate and setup hook for auto-renew
        self.add_configure_glassfish(plan)

        subjects_to_revoke = ["L=consul", "L=elastic", "L=elkadmin", "L=flink", "L=glassfish", "L=hdfs",
            "L=hive", "L=hopsmon", "L=kafka", "L=kagent", "L=livy", "L=mapred", "L=mysql", "L=rmyarn",
            "L=yarn", "L=onlinefs", "CN=hopsworks.glassfish.service"]

        self.add_revoke_old_x509_certificates(plan, subjects_to_revoke)

        # and make a REST call
        # Generate Hopsworks internal endpoint certificate
        self.add_generate_hopsworks_internal_x509(plan)

        self.add_start_consul(plan)

        # Create internal API key
        self.add_create_api_key(plan)

        # Generate Elastic signing key 
        self.add_generate_elastic_signing_key(plan)
        
        # Generate Online FS API key
        onlinefs_api_key_check = self.add_generate_onlinefs_api_key(plan)

        # Regenerate Hops host certificate
        register_generate_cert_check = self.plan_utils.get_procedure_check_file(RegisterGenerateCerts.__name__)
        
        if not self.plan_utils.check_file_exists(register_generate_cert_check) \
            and self.plan_utils.has_configured_glassfish() and not CreateServiceApiKey.should_run(self.plan_utils) \
            and self.plan_utils.check_file_exists(self.plan_utils.get_procedure_check_file(RevokeCertificates.__name__)):
            logger.info("%s - Adding procedure %s", self.name, RegisterGenerateCerts.__name__)
            p3 = RegisterGenerateCerts(RegisterGenerateCerts.__name__, self.config, register_generate_cert_check,
                                       self.cloud_provider, self.api_key, self.gateway, self.uid, self.instance_id, self.node_id, self.search_domain)
            p3.friendly_name = "Generating Hops x.509"
            p3.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Generating Hops x.509", instance_type = self.instance_type)
            plan.append(p3)

        # Setup ec2update systemd unit to callback Hopsworks Cloud when the instance has restared
        self.add_enable_ec2update(plan)

        # Configure OAuth 
        oauth_check = self.add_configure_oauth(plan)

        # Configure DAS node if there are glassfish instances in services
        # Run after all to make sure glassfish is available for services like config_oauth, service_jwt ... 
        # (Because this might restart glassfish)
        if self._has_ran_all(sed_check, register_generate_cert_check, onlinefs_api_key_check) and self._has_docker_ran() \
            and (oauth_check is None or self.plan_utils.check_file_exists(oauth_check)):
            glassfish_ha_check = self.add_configure_glassfish_ha(plan)

        # Enable and start all systemd units listed in kagent's services file
        # This should always be the last Procedure, as it signals Hopsworks Cloud
        # that the instance is up and running
        if self._has_ran_all(sed_check, register_generate_cert_check, onlinefs_api_key_check) and self._has_docker_ran() \
              and (glassfish_ha_check is None or self.plan_utils.check_file_exists(glassfish_ha_check)):
            self.add_start_enable_services(plan)

        return plan

    ## Add you checks here!
    def _has_ran_all(self, sed_check: pathlib.Path, regenerate_hops_x509_check_file: pathlib.Path, onlinefs_api_key_check: pathlib.Path) -> bool:
        return self.plan_utils.check_file_exists(sed_check) and self.plan_utils.has_issued_certificate() \
            and self.plan_utils.has_configured_glassfish() and self.plan_utils.has_started_enabled_consul() and self.plan_utils.has_enabled_ec2update_service() \
            and self.plan_utils.check_file_exists(regenerate_hops_x509_check_file) and self.plan_utils.check_file_exists(onlinefs_api_key_check)

    @abstractmethod
    def _has_docker_ran(self) -> bool:
        pass


class AWSMasterInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        # The cloud provider attribute may have already been set, but we make sure it is correct here instead
        kwargs['cloud_provider'] = cloud.Provider.AWS
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        # Setup EKS Cluster
        self.add_setup_eks(plan)

        # Setup ECR Docker images 
        self.add_setup_ecr_docker_images(plan)

        parentPlan: Sequence[Procedure] = super().create()

        return plan + list(parentPlan)

    def _has_docker_ran(self) -> bool:
        setup_ecr_docker_image_check = self.plan_utils.get_procedure_check_file(SetupECRDockerImages.__name__)
        return self.plan_utils.check_file_exists(setup_ecr_docker_image_check)


class AzureMasterInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        # The cloud provider attribute may have already been set, but we make sure it is correct here instead
        kwargs['cloud_provider'] = cloud.Provider.AZURE
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        
        self.add_setup_aks(plan)
        self.add_setup_azure_docker_images(plan)
        
        parentPlan: Sequence[Procedure] = super().create()
        return plan + list(parentPlan)
    
    def _has_docker_ran(self) -> bool:
        setup_acr_docker_image_check = self.plan_utils.get_procedure_check_file(SetupAzureDockerImages.__name__)
        return self.plan_utils.check_file_exists(setup_acr_docker_image_check)

class GCPMasterInitPlan(MasterInitPlan):
    def init(self, name: str, config: ConfigParser, **kwargs: str) -> None:
        # The cloud provider attribute may have already been set, but we make sure it is correct here instead
        kwargs['cloud_provider'] = cloud.Provider.GCP
        super().init(name, config, **kwargs)

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        
        setup_gke = self.add_setup_gke(plan)
        setup_gcr = self.add_setup_gcr_docker_images(plan)

        # Workaround to pull smarter device manager plugin image on private GKE clusters since we won't be able to pull from public registries 
        # It might be better to do that outside of ec2-init but i'm gonna leave it here for now untill we have terraform integration in place and then we can decide on this 
        if setup_gke is not None and self.plan_utils.check_file_exists(setup_gke) and setup_gcr is not None and self.plan_utils.check_file_exists(setup_gcr):
            self.add_configure_smart_device_manager(plan)

        parentPlan: Sequence[Procedure] = super().create()
        return plan + list(parentPlan)

    def _has_docker_ran(self) -> bool:
        setup_gcr_docker_image_check = self.plan_utils.get_procedure_check_file(SetupGCRDockerImages.__name__)
        return self.plan_utils.check_file_exists(setup_gcr_docker_image_check)

    def add_configure_smart_device_manager(self, plan: List[Procedure]) -> None:
        configure_smart_devices_manager_chefk = self.plan_utils.get_procedure_check_file(ConfigureSmartDeviceManagerOnGCR.__name__)
        if not self.plan_utils.check_file_exists(configure_smart_devices_manager_chefk):
            logger.info("%s - Adding procedure %s", self.name, ConfigureSmartDeviceManagerOnGCR.__name__)
            pks = ConfigureSmartDeviceManagerOnGCR(ConfigureSmartDeviceManagerOnGCR.__name__, check_file=configure_smart_devices_manager_chefk,
                                                        config=self.config, gcr_user_registry_url=self.user_registry_url, instance_id=self.instance_id)
            pks.friendly_name = "Configuring GKE"
            plan.append(pks)            
