import logging
import configparser
import subprocess

from pathlib import Path
from typing import Any, Dict, List

from ec2init import Procedure
from ec2init.plans.multinode.seds import Seds
from ec2init.plans.plan_utils import PlanUtils
from ec2init.utils import cloud
from ec2init.utils.database import databaseUtils
from ec2init.utils.hopsfs import HopsFSUtils
from ec2init.utils.services import Service, ServiceUtils
from ec2init.utils.system import SystemUtils
from ec2init.utils.cloud import get_cloud_provider
import yaml
import os 

logger = logging.getLogger(__name__)

LOGS_ENDPOINT = "{gateway}/eventsv2/ops/users/{uid}/instances/{instance_id}/logs"
METRICS_ENDPOINT = "{gateway}/eventsv2/ops/users/{uid}/instances/{instance_id}/metrics"

class MasterSeds(Seds):
    
    LOGSTASH_PIPELINES = ["jupyter", "kube_jobs", "services", "serving", "spark-streaming"]

    def __init__(self, name: str, config: configparser.ConfigParser, check_file: Path, cloud_provider: str, 
      cluster_id: str, search_domain: str, node_id: str, cluster_domain_prefix: str, gateway: str, uid: str, instance_id: str, api_key: str, deactivate_log_report: bool) -> None:
        super().__init__(name, config, cloud_provider, search_domain, node_id, gateway, uid, instance_id, api_key)
        self.check_file = check_file
        self.cluster_id = cluster_id
        self.cluster_domain_prefix = cluster_domain_prefix
        self.deactivate_log_report = deactivate_log_report
        self.local_ip = cloud.get_cloud_provider(self.cloud_provider).get_private_ip(self.config)
        self.my_secondary_node_id = PlanUtils.get_secondary_node_id(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider)

    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)
        children_procs: List[Procedure] = []

        super_procedures = super().execute()

        children_procs = super_procedures + children_procs

        public_domain_name = get_cloud_provider(self.cloud_provider)\
            .get_public_domain_name(self.config, self.cluster_id, self.cluster_domain_prefix)

        subString = "s/{0}/{1}/g".format(self.MASTER_IP, self.local_ip)

        commands: List[str] = []

        ndbprops = Seds.REPLACE.format(subString, '/srv/hops/hadoop/etc/hadoop/ndb.props')
        commands.append(ndbprops)

        epipec = Seds.REPLACE.format(subString, '/srv/hops/epipe/conf/config.ini')
        commands.append(epipec)

        epipecr = Seds.REPLACE.format(subString, '/srv/hops/epipe/conf/config-reindex.ini')
        commands.append(epipecr)

        kafkap = Seds.REPLACE.format(subString, '/srv/hops/kafka/config/server.properties')
        commands.append(kafkap)

        kafkaExternalReplacement = f"s/{self.PUBLIC_DOMAIN_NAME}/{public_domain_name}/g"
        kafkaExternal = Seds.REPLACE.format(kafkaExternalReplacement, '/srv/hops/kafka/config/server.properties')
        commands.append(kafkaExternal)

        livyc = Seds.REPLACE.format(subString, '/srv/hops/apache-livy/conf/livy.conf')
        commands.append(livyc)

        hostnameSubString = "s/{0}/{1}/g".format(self.MASTER_HOSTNAME, self.fqdn)
        kafkaph = Seds.REPLACE.format(hostnameSubString, "/srv/hops/kafka/config/server.properties")
        commands.append(kafkaph)

        elastic = Seds.REPLACE.format(hostnameSubString, "/srv/hops/opensearch/config/opensearch.yml")
        commands.append(elastic)

        runSecurityAdmin = Seds.REPLACE.format(hostnameSubString, "/srv/hops/opensearch/plugins/opensearch-security/tools/run_securityAdmin.sh")
        commands.append(runSecurityAdmin)
        
        elastic_exporter = Seds.REPLACE.format(hostnameSubString, "/lib/systemd/system/elastic_exporter.service")
        commands.append(elastic_exporter)

        logstash_url = "s;URL;{0};g".format(LOGS_ENDPOINT.format(gateway=self.gateway, uid=self.uid,instance_id=self.instance_id))
        logstash_services_conf_url = Seds.REPLACE.format(logstash_url,"/srv/hops/logstash/config/services_managed_cloud.conf")
        commands.append(logstash_services_conf_url)

        logstash_url = "s/API_KEY/{0}/g".format(self.api_key)
        logstash_services_conf_key = Seds.REPLACE.format(logstash_url,"/srv/hops/logstash/config/services_managed_cloud.conf")
        commands.append(logstash_services_conf_key)
        
        prometheus_instance_id = Seds.REPLACE.format("s/INSTANCE_ID/{0}/g".format(self.instance_id),"/srv/hops/prometheus/prometheus/prometheus.yml")
        commands.append(prometheus_instance_id)

        metrics_key_patern = "s/API_KEY/{0}/g".format(self.api_key)
        metrics_key_command = Seds.REPLACE.format(metrics_key_patern,"/srv/hops/cloud-metrics-collector/config.yml")
        commands.append(metrics_key_command)
        metrics_endpoint_patern = "s;METRICS_END_POINT;{0};g".format(METRICS_ENDPOINT.format(gateway=self.gateway, uid=self.uid,instance_id=self.instance_id))
        metrics_endpoint_command = Seds.REPLACE.format(metrics_endpoint_patern,"/srv/hops/cloud-metrics-collector/config.yml")
        commands.append(metrics_endpoint_command)

        yarn_rm_address = Seds.REPLACE.format(subString, '/srv/hops/hadoop/etc/hadoop/yarn-site.xml')
        commands.append(yarn_rm_address)

        if not self.deactivate_log_report:
            report_pattern = 's/\["services"\]/\["services","services_managed_cloud"\]/g'
            report_command = Seds.REPLACE.format(report_pattern,"/srv/hops/logstash/config/pipelines.yml")
            commands.append(report_command)
        else :
            remote_write_pattern = '/remote_write/,+5d'
            remote_write_command = Seds.REPLACE.format(remote_write_pattern,"/srv/hops/prometheus/prometheus/prometheus.yml")
            commands.append(remote_write_command)

        for c in commands:
            logger.debug("%s - Executing command %s", self.name, c)
            SystemUtils.execute_command(c.split(" "))
            logger.info("%s - Executed command %s", self.name, c)

        nn_heap_size = self._get_nn_heap_size()
        nn_heap_size_patern = "s/'-XX:MaxDirectMemorySize=3000m -Xmx3000m'/'-XX:MaxDirectMemorySize={0}m -Xmx{0}m -Xms{0}m'/g".format(nn_heap_size)
        nn_heap_size_command = Seds.REPLACE.format(nn_heap_size_patern, "/srv/hops/hadoop/sbin/set-env.sh")
        SystemUtils.execute_command([nn_heap_size_command], shell=True)

        logstash = Seds.SEEK_AND_REPLACE.format("/srv/hops/logstash/config", hostnameSubString)
        # Unix find is a pain for subprocess
        subprocess.run(logstash, timeout=20, shell=True)

        self.sed_almost_similar(self.local_ip, self.fqdn)
        
        self.setup_ha_services()

        if self.check_file:
            self.check_file.touch()

        return children_procs

    def _get_nn_heap_size(self) -> int:
        multiplier = self.config.getfloat('initplan', 'nn_heap_size_config_multiplier', fallback=0.15)
        available_memory = SystemUtils.get_system_memory()
        heap_size = int(multiplier * available_memory)
        if heap_size > available_memory:
            heap_size = available_memory
        
        heap_size = max(3000, heap_size)
        return heap_size

    def setup_ha_services(self) -> None:
        cluster_status = PlanUtils.get_cluster_status(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key)
        self.join_opensearch_cluster(cluster_status)
        self.ha_resource_managers(cluster_status)
        self.ha_zookeeper_servers(cluster_status)
        self.ha_kafka_brokers(cluster_status)
        self.ha_hive(cluster_status)
        self.ha_alertmanager()

    def join_opensearch_cluster(self, cluster_status :Dict[str, Any]) -> None:
        opensearch_config = "/srv/hops/opensearch/config/opensearch.yml"
        with open(opensearch_config, 'r') as f:
            config_dict : Dict  = yaml.load(f, Loader=yaml.loader.SafeLoader)

        all_nodes = [node["hostname"] for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Opensearch))]
        admin_node = [node["hostname"] for node in cluster_status["payload"] if node["type"] == "master"]

        config_dict['discovery.seed_hosts'] = all_nodes
        config_dict['cluster.initial_master_nodes'] = admin_node
        config_dict['plugins.security.nodes_dn'] = [f"CN={node},OU=*,L=elastic,ST=Sweden,C=SE" for node in all_nodes]
        config_dict['plugins.security.authcz.admin_dn'] = [f"CN={node},OU=0,L=elkadmin,ST=Sweden,C=SE" for node in admin_node]
        
        with open(opensearch_config, 'w') as f:
            yaml.dump(config_dict, f, sort_keys=False, default_flow_style=False)

        is_master_node_arr = [node for node in cluster_status["payload"] if node["type"] == "master" and node['ip'] == self.local_ip]
        all_nodes_ips = [node["ip"] for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Opensearch))]
        if len(is_master_node_arr) > 0:
            databaseUtils.update_variable("elastic_ip", ",".join(all_nodes_ips))

        # set multiple opensearch hosts in kibana 
        opensearch_dashboards_config_file =  "/srv/hops/opensearch-dashboards/config/opensearch_dashboards.yml"
        with open(opensearch_dashboards_config_file, 'r') as f:
            opensearch_dashboards_config : Dict  = yaml.load(f, Loader=yaml.loader.SafeLoader)
        opensearch_dashboards_config["opensearch.hosts"] = [f"https://{node}:9200" for node in all_nodes]

        with open(opensearch_dashboards_config_file, 'w') as f:
            yaml.dump(opensearch_dashboards_config, f, sort_keys=False, default_flow_style=False)

        # set multiple opensearch hosts in logstash pipelines
        open_search_hosts_str_arr = '[' + ','.join([f"\"https:\/\/{node}:9200\"" for node in all_nodes]) + ']'
        for pipeline in MasterSeds.LOGSTASH_PIPELINES:
            cmd = Seds.REPLACE.format(f"s/'hosts =>.*'/'hosts => {open_search_hosts_str_arr}'/g", f"/srv/hops/logstash/config/{pipeline}.conf")
            SystemUtils.execute_command([cmd], shell=True)
        
        # set multiple opensearch hosts in ePipe
        open_search_hosts_ip_addr = ",".join([f"{node}:9200" for node in all_nodes_ips]) 
        cmd = Seds.REPLACE.format(f"s/'elastic_addr.*'/'elastic_addr = {open_search_hosts_ip_addr}'/g", "/srv/hops/epipe/conf/config.ini")

    def ha_resource_managers(self, cluster_status :Dict[str, Any]) -> None:
        yarn_site = "/srv/hops/hadoop/etc/hadoop/yarn-site.xml"
        HopsFSUtils.update_site_xml("yarn.resourcemanager.hostname", self.local_ip, yarn_site)
        ha_resource_manager = [node for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Resourcemanager))]

        if len(ha_resource_manager) > 1:
            HopsFSUtils.update_site_xml("yarn.resourcemanager.ha.enabled", "true", yarn_site)
            HopsFSUtils.update_site_xml("yarn.resourcemanager.ha.automatic-failover.enabled", "true", yarn_site)
            rm_consul_health = "/srv/hops/hadoop/bin/consul/rm-health.sh"
            if os.path.exists(rm_consul_health):
                cmd = Seds.REPLACE.format("s/'IS_HA=\"false\"'/'IS_HA=\"true\"'/g", rm_consul_health)
                SystemUtils.execute_command([cmd], shell=True)
                cmd = Seds.REPLACE.format(f"s/'RM_ID=0'/'RM_ID={self.my_secondary_node_id}'/g", rm_consul_health)
                SystemUtils.execute_command([cmd], shell=True)

        HopsFSUtils.update_site_xml("yarn.resourcemanager.ha.id", self.my_secondary_node_id, yarn_site)
        for rm in ha_resource_manager:
            rm_id = rm["secondaryNodeId"] if "secondaryNodeId" in rm else 0
            rm_ip = rm["ip"]
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.hostname.{rm_id}", f"{rm_ip}", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.scheduler.address.{rm_id}", f"{rm_ip}:8030", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.resource-tracker.address.{rm_id}", f"{rm_ip}:8031", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.address.{rm_id}", f"{rm_ip}:8032", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.admin.address.{rm_id}", f"{rm_ip}:8033", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.webapp.address.{rm_id}", f"{rm_ip}:8088", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.webapp.https.address.{rm_id}", f"{rm_ip}:8089", yarn_site)
            HopsFSUtils.update_site_xml(f"yarn.resourcemanager.groupMembership.address.{rm_id}", f"{rm_ip}:8034", yarn_site)


        HopsFSUtils.update_site_xml("yarn.resourcemanager.ha.rm-ids",",".join([str(rm["secondaryNodeId"]) if "secondaryNodeId" in rm else "0" for rm in ha_resource_manager]), yarn_site)
        core_site = "/srv/hops/hadoop/etc/hadoop/core-site.xml"
        HopsFSUtils.update_site_xml("hadoop.proxyuser.consul.hosts", self.local_ip, core_site)
        HopsFSUtils.update_site_xml("hadoop.proxyuser.consul.users","consul", core_site)

    def ha_zookeeper_servers(self, cluster_status :Dict[str, Any]) -> None:
        # fix zk consul tag
        zk_consul_config = "/srv/hops/consul/consul.d/zk-consul.hcl"
        if os.path.exists(zk_consul_config):
            cmd = Seds.REPLACE.format(f"s/'\"1\"'/'\"0\"'/g", zk_consul_config)
            logger.debug("%s - Replace tag in zk-consul.hcl with 0", self.name)
            SystemUtils.execute_command([cmd], shell=True)

        # update configuration
        zk_config_file = "/srv/hops/zookeeper/zookeeper/conf/zoo.cfg"
        zk_servers = ["0"] + [str(node["secondaryNodeId"]) for node in cluster_status["payload"] if node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Zookeeper)]

        with open(zk_config_file, "r") as f:
                zk_config = [line for line in f.readlines() if not line.startswith("server.")]
        
        for zk_server_id in zk_servers:
            # after restart zookeeper nodes end up in cylic error state while each node tries to connect to others while the others trying to do the same 
            # Cannot open channel to x at election address
            # https://stackoverflow.com/questions/30940981/zookeeper-error-cannot-open-channel-to-x-at-election-address
            if zk_server_id == self.my_secondary_node_id:
                zk_config.append(f"server.{zk_server_id}=0.0.0.0:2888:3888\n")
            else:    
                zk_config.append(f"server.{zk_server_id}={zk_server_id}.zookeeper.service.consul:2888:3888\n")
        
        logger.info("%s - rewrite zookeeper config \n %s", self.name, "".join(zk_config))
        
        with open(zk_config_file, "w") as f:
            f.writelines(zk_config)
        
        with open("/srv/hops/zookeeper/zookeeper/data/myid", "w") as f:
            f.write(self.my_secondary_node_id)
    
    def ha_kafka_brokers(self, cluster_status :Dict[str, Any]) -> None:
        # Default broker id for kafka starts with id 1 so keep it the same for backward compabitility and to avoid issues with upgrade from 2.5.0 
        my_broker_id = int(self.my_secondary_node_id) + 1
        kafka_server_config = "/srv/hops/kafka/config/server.properties"
        cmd = Seds.REPLACE.format(f's/"broker.id=1"/"broker.id={my_broker_id}"/g', kafka_server_config)
        logger.debug("%s - Replace broker id with %s", self.name, my_broker_id)
        SystemUtils.execute_command([cmd], shell=True)

        all_nodes = [node["hostname"] for node in cluster_status["payload"] if node["type"] == "master" or (node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Kafka))]
        all_online_fs_node_ids = ["0"] + [node["secondaryNodeId"] for node in cluster_status["payload"] if node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Onlinefs)]

        all_online_fs_nodes_as_users = ";".join([f"User:{node}.onlinefs.service.consul" for node in all_online_fs_node_ids])
        all_nodes_as_users = ";".join([f"User:{node}" for node in all_nodes])
        super_users_config = f"User:kafka;User:glassfish.service.consul;{all_online_fs_nodes_as_users};{all_nodes_as_users}"

        cmd = Seds.REPLACE.format(f"s/'super.users=.*'/'super.users={super_users_config}'/", kafka_server_config)
        logger.debug("%s - Replacing super.users in %s with %s", self.name, kafka_server_config, super_users_config)
        SystemUtils.execute_command([cmd], shell=True)
    
    def ha_hive(self, cluster_status :Dict[str, Any]) -> None:
        hive_site = "/srv/hops/apache-hive/conf/hive-site.xml"
        hive_servers = [node for node in cluster_status["payload"] if node["type"] == "secondary" and ServiceUtils.containsValue(node["services"], Service.Hiveserver2)]
        if len(hive_servers) > 0:
            HopsFSUtils.update_site_xml("hive.server2.support.dynamic.service.discovery", "true", hive_site)