import logging
import configparser

from pathlib import Path
from typing import List
from ec2init.plans.multinode.seds import Seds

from ec2init import Procedure
from ec2init.plans.multinode.master_seds import MasterSeds
from ec2init.plans.plan_utils import PlanUtils
from ec2init.utils import cloud
from ec2init.utils.hopsfs import HopsFSUtils
from ec2init.utils.services import Service
from ec2init.utils.system import SystemUtils
import os 
import json 

logger = logging.getLogger(__name__)

class SecondarySeds(MasterSeds):
    def __init__(self, name: str, config: configparser.ConfigParser, check_file: Path, cloud_provider: str, 
      cluster_id: str, search_domain: str, node_id: str, cluster_domain_prefix: str, gateway: str, uid: str, instance_id: str, api_key: str, deactivate_log_report: bool, master_ip: str) -> None:
      super().__init__(name, config, check_file, cloud_provider, 
      cluster_id, search_domain, node_id, cluster_domain_prefix, gateway, uid, instance_id, api_key, deactivate_log_report)
      self.master_ip = master_ip
      self.secondary_node_id = PlanUtils.get_secondary_node_id(self.name, self.config, self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider)

    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)
        children_procs: List[Procedure] = []

        # secondary nodes doesn't have public domain setup for now, so let's use its public ip to setup external listeners in kafka
        public_ip = cloud.get_cloud_provider(self.cloud_provider).get_public_ip(self.config)
        SystemUtils.execute_command([Seds.REPLACE.format(f"s/{self.PUBLIC_DOMAIN_NAME}/{public_ip}/g", '/srv/hops/kafka/config/server.properties')], shell=True)
        
        super_procedures = super().execute()

        children_procs = super_procedures + children_procs

        self.join_consul_mesh()
        self.replace_consul_hostname_crypto_state()
        self.fix_rm_consul()
        self.fix_zk_consul()
        self.fix_glassfish_consul()

        if self.check_file:
            self.check_file.touch()

        return children_procs
    
    def join_consul_mesh(self) -> None:
        # Consul should join the master consul mesh
        subString = "s/{0}/{1}/g".format("'server = true'", f"'retry_join = [\"{self.master_ip}\"]'")
        cmd = Seds.REPLACE.format(subString, '/srv/hops/consul/consul.d/consul.hcl')
        logger.debug("%s - Removing server configuration from consul.hcl", self.name)
        SystemUtils.execute_command([cmd], shell=True)
        logger.info("%s - Removed server configuration from consul.hcl", self.name)

        subString = "s/{0}/{1}/g".format("'bootstrap_expect = 1'", "")
        cmd = Seds.REPLACE.format(subString, '/srv/hops/consul/consul.d/consul.hcl')
        logger.debug("%s - Removing bootstrap_expect configuration from consul.hcl", self.name)
        SystemUtils.execute_command([cmd], shell=True)
        logger.info("%s - Removed bootstrap_expect configuration from consul.hcl", self.name)

        subString = "s/{0}/{1}/g".format("'ui = true'", "")
        cmd = Seds.REPLACE.format(subString, '/srv/hops/consul/consul.d/consul.hcl')
        logger.debug("%s - Removing ui configuration from consul.hcl", self.name)
        SystemUtils.execute_command([cmd], shell=True)
        logger.info("%s - Removed ui configuration from consul.hcl", self.name)
        
    def replace_consul_hostname_crypto_state(self) -> None:
        kagent_conf_dir = self.config.get('general', 'kagent_conf_dir', fallback="/srv/hops/kagent/etc")
        kagent_conf_file = Path(kagent_conf_dir, 'config.ini')
        kagent_conf = configparser.ConfigParser()
        kagent_conf.read(str(kagent_conf_file))
        state_store_location = kagent_conf.get('agent', 'state-store', fallback='/srv/hops/kagent/etc/state_store')
        hopsify_state_store = os.path.join(state_store_location, "crypto_material_state.json")

        logger.debug("%s - Replacing hostname for rm and onlinefs in %s", self.name, str(hopsify_state_store))
        with open(hopsify_state_store, 'r') as fd:
            cms = json.load(fd)

        for s in cms["crypto_state_store"]:
            if s == "onlinefs":
                cms["crypto_state_store"][s]["common_name"] = f"{self.secondary_node_id}.onlinefs.service.consul"

        with open(hopsify_state_store, 'w') as fd:
            json.dump(cms, fd)

        logger.info("%s - Replaced hostname for rm and onlinefs in %s", self.name, str(hopsify_state_store))

    def fix_rm_consul(self) -> None:
        rm_consul_config = '/srv/hops/consul/consul.d/rm-consul.hcl'
        if os.path.exists(rm_consul_config):
            cmd = Seds.REPLACE.format(f"s/'\"0\"'/'\"{self.secondary_node_id}\"'/g", rm_consul_config)
            logger.debug("%s - Replace 0 tag in rm-consul.hcl with %s", self.name, self.secondary_node_id)
            SystemUtils.execute_command([cmd], shell=True)

    
    def fix_zk_consul(self) -> None:
        zk_consul_config = "/srv/hops/consul/consul.d/zk-consul.hcl"
        if os.path.exists(zk_consul_config):
            cmd = Seds.REPLACE.format(f"s/'\"0\"'/'\"{self.secondary_node_id}\"'/g", zk_consul_config)
            logger.debug("%s - Replace tag in zk-consul.hcl with %s", self.name, self.secondary_node_id)
            SystemUtils.execute_command([cmd], shell=True)

    def fix_glassfish_consul(self)-> None:
        glassfish_consul_config = "/srv/hops/consul/consul.d/glassfish-consul.hcl"
        if os.path.exists(glassfish_consul_config):
            glassfish_worker_consul_hcl='''services {
    id = "hopsworks/web"
    name = "glassfish"
    tags = [
        "http",
        "monitoring",
        "hopsworks"
    ]
    port = 8182
    check = {
        id = "hw-health-check"
        name = "Health check for Hopsworks"
        args = ["/srv/hops/domains/domain1/bin/glassfish-health.sh"]
        interval = "20s"
        timeout = "20s"
        status = "passing"
    }
}'''
            with open(glassfish_consul_config, 'w') as file:
                file.write(glassfish_worker_consul_hcl)
            logger.debug("%s - Replace glassfish-consul.hcl with glassfish-worker-consul.hcl", self.name)