
import logging
from pathlib import Path
from typing import Any, Dict, List

from ec2init.execution.procedure import Procedure
from ec2init.utils.database import databaseUtils
from ec2init.utils.system import SystemUtils


logger = logging.getLogger(__name__)

class UpgradeOAuthConfig(Procedure):
    def __init__(self, name: str, check_file: Path, oauth_conf: Dict[str, Any], user_management_type: str) -> None:
        self.name = name
        self.check_file = check_file
        self.oauth_conf = oauth_conf
        self.user_management_type = user_management_type

    def pre(self) -> None:
        if self.report_callback is not None:
            try:
                self.report_callback.execute()
            except Exception as ex:
                logger.warning("%s - Could not execute pre-execute: %s", self.name, ex)
    
    def execute(self) -> List[Procedure]:
        if self.oauth_conf:
            logger.info("%s - Starting execution", self.name)

            oauth_enabled = SystemUtils.execute_mysql_statement("SELECT value from variables WHERE id='oauth_enabled'")[0][0].upper() == "TRUE"
            if oauth_enabled and self.user_management_type.upper() == "MANAGED":
                logger.info("managed user")
                # It is a cluster with usermanagement enabled.

                # Extract the information about the managed user group mapping.
                expected_mapping_string = self.oauth_conf['groupMappings']
                expected_mapping_parts = expected_mapping_string.split(";")
                expected_mapping = {}
                for part in expected_mapping_parts:
                    id = part.split('->')[0]
                    expected_mapping[id]=part

                # Get the currently configured mapping. We don't want to overwrite any mapping added by the user.
                group_mapping = SystemUtils.execute_mysql_statement("SELECT value from variables WHERE id='oauth_group_mapping'")[0][0]

                logger.info("group mapping %s", group_mapping)
                # Check if all the mapping for managed user are present, add the missing ones.
                count = 0
                for id in expected_mapping:
                    if id not in group_mapping:
                        group_mapping = group_mapping + ";" + expected_mapping[id]
                        count = count+1

                # If managed user was set before upgrade only one mapping should be missing and need to be added.
                # If more than one mapping is missing it means that the user removed the user management manually
                # we do not want to reenable it without the user knowing it.
                if count == 1:
                    logger.info("updating group mapping %s", group_mapping)
                    databaseUtils.update_variable("oauth_group_mapping", group_mapping)


            if self.check_file is not None:
                self.check_file.touch()
            logger.info("%s - Finished execution", self.name)
        return []

    def rollback(self) -> None:
        pass