import logging
import configparser

from typing import List

from ec2init import Procedure
from ec2init.utils.system import SystemUtils
from ec2init.plans.multinode.seds import Seds
from ec2init.utils.systemd import SystemdUtils

logger = logging.getLogger(__name__)

HEARTBEAT_ENDPOINT = "{gateway}/eventsv2/ndbagent/{uid}/{cid}/heartbeat"


class StartNdbAgent(Procedure):
    AGENT_CONFIG = "/srv/hops/ndb-agent/ndb-agent/config/config.yml"

    def __init__(
        self,
        name: str,
        config: configparser.ConfigParser,
        api_key: str,
        is_bootstrap: bool,
        gateway: str,
        uid: str,
        cluster_id: str,
        extra_mysqld_vms: str,
    ) -> None:
        self.name = name
        self.config = config
        self.api_key = api_key
        self.is_bootstrap = is_bootstrap
        self.hb_endpoint = HEARTBEAT_ENDPOINT.format(
            gateway=gateway, uid=uid, cid=cluster_id
        )
        self.extra_mysqld_vms = extra_mysqld_vms

    def execute(self) -> List[Procedure]:
        logger.debug(f"{self.name} - Replacing API key")
        subString = f"s/API_KEY/{self.api_key}/g"
        cmd = Seds.REPLACE.format(subString, StartNdbAgent.AGENT_CONFIG)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info(f"{self.name} - Replaced API key")

        logger.debug(f"{self.name} - Configure extra_mysqld_vms")
        subString = (
            f"s/'extra-mysqld-vms:.*'/'extra-mysqld-vms: {self.extra_mysqld_vms}'/g"
        )
        cmd = Seds.REPLACE.format(subString, StartNdbAgent.AGENT_CONFIG)
        SystemUtils.execute_command([cmd], shell=True)

        logger.debug(f"{self.name} - Replacing Heartbeat endpoint")
        subString = f"s;HEARTBEAT_ENDPOINT;{str(self.hb_endpoint)};g"
        cmd = Seds.REPLACE.format(subString, StartNdbAgent.AGENT_CONFIG)
        SystemUtils.execute_command(cmd.split(" "))
        logger.info(f"{self.name} - Replaced Heartbeat endpoint")
        
        if self.is_bootstrap:
            logger.debug(f"{self.name} - Enabling Heartbeat")
            subString = f"s/HEARTBEAT_ENABLED/true/g"
            cmd = Seds.REPLACE.format(subString, StartNdbAgent.AGENT_CONFIG)
            SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Enabled Heartbeat")
        else:
            logger.debug(f"{self.name} - Disabling Heartbeat")
            subString = f"s/HEARTBEAT_ENABLED/false/g"
            cmd = Seds.REPLACE.format(subString, StartNdbAgent.AGENT_CONFIG)
            SystemUtils.execute_command(cmd.split(" "))
            logger.info(f"{self.name} - Disabled Heartbeat")

        logger.debug(f"{self.name} - Starting ndb-agent")
        SystemUtils.execute_command(["systemctl", "enable", "ndb-agent"], timeout=120)
        SystemUtils.execute_command(["systemctl", "start", "ndb-agent"], timeout=120)
        logger.info(f"{self.name} - Started ndb-agent")

        return []

    def rollback(self):
        pass

    @staticmethod
    def has_run() -> bool:
        return SystemdUtils.is_active("ndb-agent") and SystemdUtils.is_enabled("ndb-agent")
