import logging
from configparser import ConfigParser
from pathlib import Path

from typing import List, Set

from ec2init import Procedure
from ec2init.utils.cloud import get_cloud_provider
from ec2init.utils.system import SystemUtils
from ec2init.plans.plan_utils import HopsPlanUtils
from ec2init.plans.multinode.seds import Seds
import os 
import time
from ec2init.utils.shell import ShellUtils
import subprocess
from ec2init.utils.versions import VersionUtils

logger = logging.getLogger(__name__)

class Expat(Procedure):
    def __init__(self, name: str, config: ConfigParser, check_file: Path, source_version: str, upgrade_version: str, cloud_provider: str, search_domain: str, node_id: str, default_registry_url: str, python_image_name: str, managed_containers: bool):
        self.name = name
        self.config = config
        self.check_file = check_file
        self.upgrade_version_unclean = upgrade_version
        self.source_version = HopsPlanUtils.clean_hopsworks_version(source_version)
        self.upgrade_version = HopsPlanUtils.clean_hopsworks_version(upgrade_version)
        self.cloud_provider = cloud_provider
        self.search_domain = search_domain
        self.node_id = node_id
        self.default_registry_url = default_registry_url
        self.python_image_name = python_image_name
        self.base_image_name = "base"
        self.managed_containers = managed_containers
        self.expat_dir = self.config.get('upgradeplan', 'expat_dir', fallback='/root/expat')
        self.hopsworks_versions = self.config.get('upgradeplan', 'hopsworks_versions').split(',')
        self.tours_cache_base_dir = Path(self.config.get('upgradeplan', 'tours_cache_base_dir', fallback='/home/hdfs/tours_cache'))
        self.tours_cache_info_csv = Path(self.config.get('upgradeplan', 'tours_cache_info_csv', fallback='tours_info.csv'))
        self.tours_rollback_cache_info_csv = Path(self.tours_cache_base_dir, "tours_rollback_info.csv")
        self.old_tour_files_cache_on_hopsfs = "/.old_tours_files"
        self.expat_logs_dir = f"{self.expat_dir}/logs"
        self.source_version_with_zero_patch = VersionUtils.get_version_with_zero_patch_version(self.source_version)

    def pre(self) -> None:
        pass
    
    def execute(self) -> List[Procedure]:
        logger.info("%s - Starting execution", self.name)
        try:
            if not self.managed_containers:
                # Push base and python images to local registry
                base_image_process = subprocess.Popen(["docker", "push", f"{self.default_registry_url}/{self.base_image_name}:{self.upgrade_version_unclean}"])
                python_image_proccess = subprocess.Popen(["docker", "push", f"{self.default_registry_url}/{self.python_image_name}:{self.upgrade_version_unclean}"])

            self._fix_expat_config()
            self._run_expat_migrate()
            
            # wait for docker push to finish
            if not self.managed_containers:
                base_image_process.wait()
                python_image_proccess.wait()

            self._add_new_tour_files()

            # Every thing is OK.
            if self.check_file is not None:
                self.check_file.touch()
        
        except Exception as ex:
            logger.warning("%s - Could not run Expat.", ex)
            raise ex

        logger.info("%s - Finished my task!", self.name)
        return []
    
    def rollback(self) -> None:
        self._rollback_old_tour_files()
        self._run_expat_rollback()

    def _fix_expat_config(self) -> None:
        if os.path.exists(self.expat_dir):
            fqdn = get_cloud_provider(self.cloud_provider).get_fqdn(self.config, self.search_domain, self.node_id)
            expat_config = f"{self.expat_dir}/etc/expat-site.xml"
            SystemUtils.execute_command(Seds.REPLACE.format("s/ELASTIC_URL/https\:\/\/{0}\:9200/g".format(fqdn), expat_config).split(" "))
            SystemUtils.execute_command(Seds.REPLACE.format("s/KIBANA_URL/https\:\/\/{0}\:5601/g".format(fqdn), expat_config).split(" "))
            if not os.path.exists(self.expat_logs_dir):
                os.mkdir(self.expat_logs_dir)

    def _run_expat_migrate(self) -> None:
        if os.path.exists(self.expat_dir):
            run_expat_script = f"{self.expat_dir}/run_expat.sh"
            start_expat = False
            for version in self.hopsworks_versions:
                if self.source_version_with_zero_patch == version:
                    start_expat = True
                    continue

                if start_expat:
                    SystemUtils.execute_command([f"{run_expat_script} migrate {version} > {self.expat_logs_dir}/migrate_{version}_{time.time()} 2>&1"], shell=True, timeout=None)
                    
                if self.upgrade_version == version:
                    break
        else:
            logger.info("Skipping %s migrate since %s doesn't exists", self.name, self.expat_dir)
    
    def _run_expat_rollback(self) -> None:
        if os.path.exists(self.expat_dir):
            run_expat_script = f"{self.expat_dir}/run_expat.sh"
            reversed_hopsworks_versions = self.hopsworks_versions[::-1]
            for version in reversed_hopsworks_versions:
                if self.source_version_with_zero_patch == version:
                    break
                SystemUtils.execute_command([f"{run_expat_script} rollback {version} > {self.expat_logs_dir}/rollback_{version}_{time.time()} 2>&1"], shell=True, timeout=None)
        else:
            logger.info("Skipping %s migrate since %s doesn't exists", self.name, self.expat_dir)

    def _add_new_tour_files(self) -> None:
        tours_info_csv : Path = Path(self.tours_cache_base_dir, self.tours_cache_info_csv)
        # Move required tour files to hdfs
        if tours_info_csv.exists():
            logger.info("%s - Copy tour files after upgrade", self.name)
            tour_files: Set[str] = set()
            with open(tours_info_csv, "r") as file:
                tour_files = set(file.readlines())
            
            try:
                SystemUtils.execute_hdfs_command(f"dfs -rm -r -f {self.old_tour_files_cache_on_hopsfs}")
            except Exception as ex:
                logger.debug("Skip exception %s when removing old tour files cache", ex)

            rollback_tour_files = []
            SystemUtils.execute_hdfs_command(f"dfs -mkdir {self.old_tour_files_cache_on_hopsfs}")
            for tour in tour_files:
                data_points = tour.strip().split(',')
                if len(data_points) == 5:
                    src, dst, owner, group, mode = data_points
                    dst_tmp = f"{self.old_tour_files_cache_on_hopsfs}/{os.path.basename(dst)}"
                    try:
                        SystemUtils.execute_hdfs_command(f"dfs -mv {dst} {dst_tmp}")
                        rollback_tour_files.append(f"add,{dst_tmp},{dst}")
                    except Exception as ex:
                        logger.info("Skip exception [%s] when moving %s to tmp directory %s", ex, dst, dst_tmp)
                        rollback_tour_files.append(f"del,{dst_tmp},{dst}")
                    SystemUtils.execute_hdfs_command(f"dfs -copyFromLocal {src} {dst}")
                    SystemUtils.execute_hdfs_command(f"dfs -chown -R {owner} {dst}")
                    SystemUtils.execute_hdfs_command(f"dfs -chgrp -R {group} {dst}")
                    SystemUtils.execute_hdfs_command(f"dfs -chmod -R {mode} {dst}")
                    
        self.tours_rollback_cache_info_csv.write_text("\n".join(rollback_tour_files))

        if self.tours_cache_base_dir.exists():
            ShellUtils.rmtree(str(self.tours_cache_base_dir.resolve()))

    def _rollback_old_tour_files(self) -> None:
        if self.tours_rollback_cache_info_csv.exists():
            tour_files: List[str] = []
            with open(self.tours_rollback_cache_info_csv, "r") as file:
                tour_files = file.readlines()
            for tour in tour_files:
                data_points = tour.strip().split(',')
                if len(data_points) == 3:
                    op, src, dst = data_points
                    SystemUtils.execute_hdfs_command(f"dfs -rm -r -f {dst}")
                    if op == "add":
                        SystemUtils.execute_hdfs_command(f"dfs -mv {src} {dst}")
            
            SystemUtils.execute_hdfs_command(f"dfs -rm -r -f {self.old_tour_files_cache_on_hopsfs}")