import logging
import configparser
import os
import shutil

from typing import Sequence, List

from ec2init import Procedure
from ec2init.plans.init.base_init_plan import BaseInitPlan, MasterBaseInitPlan
from ec2init.plans.preinit.format_nvme_drives import WorkerFormatNVMEDrives
from ec2init.plans.preinit.pre_inits import MasterPreInit, SecondaryPreInit, WorkerPreInit, MysqlNodePreInit
from ec2init.plans.preinit.database_init import MasterInitializeDatabase, MasterWithNdbInitializeDatabase, SecondaryWithMySQLd
from ec2init.plans.preinit.grow_data_volume import FixBackupDataVolume, GrowDataVolume
from ec2init.plans.init.check_s3_permissions import CheckS3Permissions
from ec2init.utils.offline_backup_information import OffLineBackupInformation
from ec2init.utils.cloud import Provider
from ec2init.plans.init.check_azure_container_permissions import CheckAzureContainerPermissions
from ec2init.plans.init.check_gcp_bucket_permissions import CheckGCPBucketPermissions
from ec2init.plans.init.setup_ecr_docker_images import SetupECRDockerImages

logger = logging.getLogger(__name__)

class FixBackupDataVolumePlan(MasterBaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "master"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        self.add_fix_backup_data_volume(plan)
        return plan


    def add_fix_backup_data_volume(self, plan: List[Procedure]) -> None:
        check_file = self.plan_utils.get_procedure_check_file(FixBackupDataVolume.__name__)
        if not self.plan_utils.check_file_exists(check_file):
            logger.info(f"{self.name} - Adding procedure {FixBackupDataVolume.__name__}")
            fix_vol = FixBackupDataVolume(FixBackupDataVolume.__name__, self.config, self.plan_utils)
            fix_vol.friendly_name = "Fixing backup data volume mount point"
            fix_vol.fail_plan = True
            plan.append(fix_vol)

class MasterPreInitPlan(MasterBaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.user_mgm_type = kwargs.get("user_management_type", "")
        self.create_pwd_user = False if kwargs.get("create_pwd_user", "").lower() == "false" else True
        if OffLineBackupInformation.is_backup(self.offline_backup_info):
            # We are restoring a multidisk backup
            # we need to delete some more check files
            if self.plan_utils.check_file_exists(self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)):
                files_to_keep = [
                    GrowDataVolume.__name__,
                    "backup_info"
                ]
                procedures_check_directory = self.plan_utils.get_procedures_check_directory()
                for file in os.listdir(procedures_check_directory):
                    if file in files_to_keep:
                        continue
                    full_path = os.path.join(procedures_check_directory, file)
                    logger.info(f"Deleting file {full_path}")
                    if os.path.isfile(full_path):
                        os.remove(full_path)
                    else:
                        shutil.rmtree(full_path)
    
    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        grow_data_vol_check = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        init_db_check_file = self.plan_utils.get_procedure_check_file(MasterInitializeDatabase.__name__)
        if not self.plan_utils.check_file_exists(init_db_check_file)\
            and self.plan_utils.check_file_exists(grow_data_vol_check):
            if self.mgmd_ip == "":
                logger.info("%s - Adding procedure %s", self.name, MasterInitializeDatabase.__name__)
                init_db_task = MasterInitializeDatabase(MasterInitializeDatabase.__name__, self.config, init_db_check_file,
                    self.bucket_configuration, self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email,
                    self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider, self.is_upgrade, self.search_domain, self.node_id, self.user_mgm_type, 
                    self.create_pwd_user, self.source_version, self.is_backup, self.offline_backup_info, self.cluster_domain_prefix)
                init_db_task.friendly_name = "Initializing database"
            else:
                logger.info("%s - Adding procedure %s", self.name, MasterWithNdbInitializeDatabase.__name__)
                ## Check file intentionally is the same for both Procedures
                init_db_task = MasterWithNdbInitializeDatabase(MasterWithNdbInitializeDatabase.__name__, self.config, init_db_check_file,
                    self.bucket_configuration, self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email,
                    self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider, self.mgmd_ip, self.is_upgrade, self.search_domain, 
                    self.node_id, self.user_mgm_type, self.create_pwd_user, self.source_version, self.is_backup, self.offline_backup_info, self.cluster_domain_prefix)
                init_db_task.friendly_name = "Configuring the head node connection to the database"
            init_db_task.fail_plan = True
            init_db_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Initializing database", instance_type = self.instance_type)
            plan.append(init_db_task)

        preinit_check_file = self.plan_utils.get_procedure_check_file(MasterPreInit.__name__)
        if not self.plan_utils.check_file_exists(preinit_check_file)\
            and self.plan_utils.check_file_exists(init_db_check_file):
            logger.info("%s - Adding procedure %s", self.name, MasterPreInit.__name__)
            preinit_task = MasterPreInit(MasterPreInit.__name__, self.config, preinit_check_file, self.bucket_configuration,
                self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email, self.gateway, self.uid,
                self.instance_id, self.api_key, self.cloud_provider, self.user_mgm_type, self.create_pwd_user, self.is_upgrade, self.source_version, self.version, 
                self.search_domain, self.node_id, self.is_backup, self.offline_backup_info, self.cluster_domain_prefix, self.services_list, k8s_enabled=bool(self.k8s_cluster_name), enable_flyingduck=self.enable_flyingduck, nlb_domain=self.nlb_domain)
            preinit_task.friendly_name = "Master node pre init"
            preinit_task.fail_plan = True
            preinit_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing instance", instance_type = self.instance_type)
            plan.append(preinit_task)

        return plan

class WorkerPreInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "worker"

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        maybe_format_nvme_drives_check = self.plan_utils.get_procedure_check_file(WorkerFormatNVMEDrives.__name__)
        if not self.plan_utils.check_file_exists(maybe_format_nvme_drives_check):
            logger.info(f"{self.name} - Adding procedure {WorkerFormatNVMEDrives.__name__}")
            t = WorkerFormatNVMEDrives(WorkerFormatNVMEDrives.__name__, maybe_format_nvme_drives_check)
            plan.append(t)

        preinit_check_file = self.plan_utils.get_procedure_check_file(WorkerPreInitPlan.__name__)
        if not self.plan_utils.check_file_exists(preinit_check_file):
            logger.info("%s - Adding procedure %s", self.name, WorkerPreInitPlan.__name__)
            worker_preinit_task = WorkerPreInit(WorkerPreInit.__name__, self.config, preinit_check_file,
                self.bucket_configuration, self.region, self.api_key, self.cloud_provider, self.search_domain, self.node_id)
            worker_preinit_task.friendly_name = "Worker node pre init"
            worker_preinit_task.fail_plan = True
            plan.append(worker_preinit_task)
        return plan

class RonDBPreInitPlan(BaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)

    def add_s3_check(self, plan: List[Procedure]) -> None:
        # Check S3 bucket permissions
        if self.bucket_configuration.has_bucket() and not OffLineBackupInformation.is_backup(self.offline_backup_info):
            s3_check = self.plan_utils.get_procedure_check_file(
                CheckS3Permissions.__name__)
            s3_test_obj = self.plan_utils.get_cloud_tmp_test_file()

            if not self.plan_utils.check_file_exists(s3_check):
                logger.info("%s - Adding procedure %s", self.name,
                            CheckS3Permissions.__name__)
                s3p = CheckS3Permissions(CheckS3Permissions.__name__, check_file=s3_check, bucket_configuration=self.bucket_configuration,\
                    region=self.region, s3_test_obj=s3_test_obj, is_backup_or_upgrade=self.is_upgrade or self.is_backup,\
                    deleted_blocks_retention_days=self.deleted_blocks_retention_days)
                s3p.friendly_name = "Checking S3 bucket is empty and the instance has read,write,list and delete privileges to the bucket"
                s3p.fail_plan = True
                s3p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Checking S3 permissions", instance_type=self.get_instance_type())
                plan.append(s3p)
        else:
            logger.info("Skipping S3 checks. No S3 bucket specified")

    def add_azure_container_checks(self, plan: List[Procedure]) -> None:

        # Test Azure container connection string and Permissions
        if self.bucket_configuration.has_bucket() and not OffLineBackupInformation.is_backup(self.offline_backup_info):
            container_check = self.plan_utils.get_procedure_check_file(
                CheckAzureContainerPermissions.__name__)
            azure_test_obj = self.plan_utils.get_cloud_tmp_test_file()

            if not self.plan_utils.check_file_exists(container_check):
                logger.info("%s - Adding procedure %s", self.name,
                            CheckAzureContainerPermissions.__name__)
                p = CheckAzureContainerPermissions(CheckAzureContainerPermissions.__name__, check_file=container_check,
                                        container_configuration=self.bucket_configuration, region=self.region, test_obj=azure_test_obj,
                                        storage_account=self.azure_storage_account, client_id=self.azure_managed_identity_id,
                                        deleted_blocks_retention_days=self.deleted_blocks_retention_days, is_backup_or_upgrade=self.is_upgrade or self.is_backup,
                                        azure_subscription_id=self.azure_subscription_id, azure_resource_group=self.azure_resource_group)
                p.friendly_name = "Checking Azure Blob Storage permissions"
                p.fail_plan = True
                p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Checking Azure Blob Storage permissions", instance_type=self.get_instance_type())
                plan.append(p)
        else:
            logger.info("Skipping Azure container checks. No container specified")

    def add_gcp_bucket_checks(self, plan: List[Procedure]) -> None:
        if self.bucket_configuration.has_bucket() and not OffLineBackupInformation.is_backup(self.offline_backup_info):
            bucket_check = self.plan_utils.get_procedure_check_file(CheckGCPBucketPermissions.__name__)
            gcp_test_obj = self.plan_utils.get_cloud_tmp_test_file()
            if not self.plan_utils.check_file_exists(bucket_check):
                logger.info("%s - Adding procedure %s", self.name, CheckGCPBucketPermissions.__name__)
                p= CheckGCPBucketPermissions(CheckGCPBucketPermissions.__name__, check_file=bucket_check, bucket_configuration=self.bucket_configuration, zone = self.zone, deleted_blocks_retention_days=self.deleted_blocks_retention_days,
                                    is_backup_or_upgrade=self.is_upgrade or self.is_backup, test_obj=gcp_test_obj, config=self.config, service_account=self.gcp_service_account)
                p.friendly_name = "Checking Bucket permissions"
                p.fail_plan = True
                p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Checking Bucket permissions", instance_type=self.get_instance_type())
                plan.append(p)
        else:
            logger.info("Skipping GCP container checks. No container specified")

    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)
        self.add_callback_instance_booted(plan, user_email=self.user_email)
        self.add_grow_data_volume(plan)
        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)
        return plan

class RonDBAllInOnePreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "rondb_all-in-one"

    def create(self) -> Sequence[Procedure]:
        parentPlan: Sequence[Procedure] = super().create()
        plan: List[Procedure] = list(parentPlan)
        # Check bucket permissions
        if self.cloud_provider == Provider.AWS:
            self.add_s3_check(plan)
        if self.cloud_provider == Provider.AZURE:
            # Check for Azure Block storage permissions
            self.add_azure_container_checks(plan)
        if self.cloud_provider == Provider.GCP:
            self.add_gcp_bucket_checks(plan)

        # Setup Docker images this is not really needed for this type of node 
        # but this allow us to check early that we have access to the repository
        if self.cloud_provider == Provider.AWS:
            self.add_setup_ecr_docker_images(plan)
        elif self.cloud_provider == Provider.AZURE:
            self.add_setup_azure_docker_images(plan)
        elif self.cloud_provider == Provider.GCP:
            self.add_setup_gcr_docker_images(plan)

        return plan

class MgmRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_mgm"

    def create(self) -> Sequence[Procedure]:
        parentPlan: Sequence[Procedure] = super().create()
        plan: List[Procedure] = list(parentPlan)
        # Check bucket permissions
        if self.cloud_provider == Provider.AWS:
            self.add_s3_check(plan)
        if self.cloud_provider == Provider.AZURE:
            # Check for Azure Block storage permissions
            self.add_azure_container_checks(plan)
        if self.cloud_provider == Provider.GCP:
            self.add_gcp_bucket_checks(plan)

        # Setup Docker images this is not really needed for this type of node 
        # but this allow us to check early that we have access to the repository
        if self.cloud_provider == Provider.AWS:
            self.add_setup_ecr_docker_images(plan)
        elif self.cloud_provider == Provider.AZURE:
            self.add_setup_azure_docker_images(plan)
        elif self.cloud_provider == Provider.GCP:
            self.add_setup_gcr_docker_images(plan)

        return plan

class NdbdRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_ndbd"

class MysqldRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "mysqld"
    
    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []
        self.add_mysqld_preinit(plan)
        
        if self.enable_flyingduck:
            if self.cloud_provider == Provider.AWS:
                self.add_setup_ecr_docker_images(plan)
            elif self.cloud_provider == Provider.AZURE:
                self.add_setup_azure_docker_images(plan)
            elif self.cloud_provider == Provider.GCP:
                self.add_setup_gcr_docker_images(plan)
        
        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)
        return plan
    
    def add_setup_ecr_docker_images(self, plan: List[Procedure]) -> None:
        self.add_setup_ecr_docker_images_int(plan, self.enable_flyingduck)

    def add_mysqld_preinit(self, plan: List[Procedure]) -> None:
        mysqld_preint_check = self.plan_utils.get_procedure_check_file(MysqlNodePreInit.__name__)
        if not self.plan_utils.check_file_exists(mysqld_preint_check):
            logger.info("%s - Adding procedure %s", self.name, MysqlNodePreInit.__name__)
            p= MysqlNodePreInit(MysqlNodePreInit.__name__, mysqld_preint_check, self.config, self.enable_flyingduck, self.docker_repos, self.cloud_provider, self.search_domain, self.node_id)
            p.friendly_name = "Configuring MySQL node"
            p.fail_plan = True
            p.report_callback = self.plan_utils.create_callback_procedure(
                    "starting", stage="Configuring MySQL node", instance_type=self.get_instance_type())
            plan.append(p)

class ApiRonDBPreInitPlan(RonDBPreInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        self.instance_type = "ndb_api"

class SecondaryPreInitPlan(MasterBaseInitPlan):
    def init(self, name: str, config: configparser.ConfigParser, **kwargs: str) -> None:
        super().init(name, config, **kwargs)
        # TODO handle backups?!!
        self.instance_type = "secondary"
    
    def create(self) -> Sequence[Procedure]:
        plan: List[Procedure] = []

        self.add_centos_configure_dns(plan)

        # Make a callback to Hopsworks Cloud indicating the instance has booted
        self.add_callback_instance_booted(plan)

        parentPlan: Sequence[Procedure] = super().create()
        plan += list(parentPlan)

        grow_data_vol_check = self.plan_utils.get_procedure_check_file(GrowDataVolume.__name__)
        init_db_check_file = self.plan_utils.get_procedure_check_file(SecondaryWithMySQLd.__name__)
        if not self.plan_utils.check_file_exists(init_db_check_file)\
            and self.plan_utils.check_file_exists(grow_data_vol_check):
            logger.info("%s - Adding procedure %s", self.name, SecondaryWithMySQLd.__name__)
            init_db_task = SecondaryWithMySQLd(SecondaryWithMySQLd.__name__, self.config, init_db_check_file,
                    self.bucket_configuration, self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email,
                    self.gateway, self.uid, self.instance_id, self.api_key, self.cloud_provider,
                    self.mgmd_ip, self.is_upgrade, self.search_domain, self.node_id, self.user_mgm_type, self.create_pwd_user, self.source_version, self.is_backup, self.offline_backup_info, self.cluster_domain_prefix)
            init_db_task.friendly_name = "Secondary Head - initializing mysqld"
            init_db_task.fail_plan = True
            init_db_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Initializing mysqld", instance_type = self.instance_type)
            plan.append(init_db_task)

        preinit_check_file = self.plan_utils.get_procedure_check_file(SecondaryPreInit.__name__)
        if not self.plan_utils.check_file_exists(preinit_check_file)\
            and self.plan_utils.check_file_exists(init_db_check_file):
            logger.info("%s - Adding procedure %s", self.name, SecondaryPreInit.__name__)
            preinit_task = SecondaryPreInit(SecondaryPreInit.__name__, self.config, preinit_check_file, self.bucket_configuration,
                self.region, self.ldap_conf_raw, self.oauth_conf_raw, self.password_file, self.user_email, self.gateway, self.uid,
                self.instance_id, self.api_key, self.cloud_provider, self.user_mgm_type, self.create_pwd_user, self.is_upgrade, self.source_version, self.version, self.search_domain, self.node_id, self.is_backup, self.offline_backup_info,
                self.cluster_domain_prefix, self.services_list, k8s_enabled=bool(self.k8s_cluster_name), enable_flyingduck=self.enable_flyingduck, nlb_domain=self.nlb_domain)
            preinit_task.friendly_name = "Secondary node pre init"
            preinit_task.fail_plan = True
            preinit_task.report_callback = self.plan_utils.create_callback_procedure("starting", stage="Preparing instance", instance_type = self.instance_type)
            plan.append(preinit_task)

        return plan