from typing import Any, Optional

from ec2init.utils.cloud import Provider

class BucketEncryptionConfiguration:

    def __init__(self, encryption_mode: str) -> None:
        self.encryption_mode = encryption_mode

    def is_encryption_enabled(self) -> bool:
        return self.encryption_mode.upper() != "NONE"

class S3BucketEncryptionConfiguration(BucketEncryptionConfiguration):
    
    def __init__(self, encryption_mode: str, kms_type: str, user_key_arn: str, bucket_key: bool) -> None:
        super().__init__(encryption_mode)
        self.kms_type = kms_type
        self.user_key_arn = user_key_arn
        self.bucket_key = bucket_key

    def is_kms_enabled(self) -> bool:
        return self.encryption_mode.upper() == "SSE-KMS"

    def is_managed_kms(self) -> bool:
        return self.kms_type.upper() == "MANAGED"

class BucketGeneralConfiguration:
    pass

class S3GeneralConfiguration(BucketGeneralConfiguration):
    @property
    def acl_bucket_full_control(self) -> bool:
        return self._acl_bucket_full_control

    @acl_bucket_full_control.setter
    def acl_bucket_full_control(self, enabled: bool) -> None:
        self._acl_bucket_full_control = enabled
        
class BucketConfiguration:
    def __init__(self, name: str, encryption_configuration: BucketEncryptionConfiguration, general_configuration: Optional[BucketGeneralConfiguration] = None) -> None:
        self.name = name
        self.encryption_configuration = encryption_configuration
        self.general_configuration = general_configuration

    def has_bucket(self) -> bool:
        return self.name is not None and self.name != ""

    @staticmethod
    def get_bucket_configuration(bucket_cloud_provider: str, **kwargs) -> 'BucketConfiguration':
        bucket = kwargs.get("bucket", "")
        if bucket_cloud_provider.lower() == Provider.AWS:
            return BucketConfiguration._get_s3_bucket_configuration(bucket, **kwargs)

        if bucket_cloud_provider.lower() == Provider.AZURE:
            return BucketConfiguration._get_azure_bucket_configuration(bucket, **kwargs)

        if bucket_cloud_provider.lower() == Provider.GCP:
            return BucketConfiguration._get_gcs_bucket_configuration(bucket, **kwargs)

        raise Exception(f"unknown provider {bucket_cloud_provider}")

    @staticmethod
    def _get_s3_bucket_configuration(name: str, **kwargs) -> 'BucketConfiguration':
        encryption_mode: str = kwargs.get("bucket_encryption_mode", "None")
        kms_bucket_key: bool = kwargs.get("bucket_kms_bucket_key", "true").lower() == "true"
        kms_type: str = kwargs.get("bucket_kms_type", "")
        kms_user_key_arn: str = kwargs.get("bucket_kms_user_key_arn", "")
        s3_acl_owner_full_control: bool = kwargs.get("s3_acl_owner_full_control", "false").lower() == "true"
        general_conf = S3GeneralConfiguration()
        general_conf.acl_bucket_full_control = s3_acl_owner_full_control

        return BucketConfiguration(name, S3BucketEncryptionConfiguration(encryption_mode, kms_type, kms_user_key_arn, kms_bucket_key), general_conf)

    @staticmethod
    def _get_azure_bucket_configuration(name: str, **kwargs) -> 'BucketConfiguration':
        return BucketConfiguration(name, BucketEncryptionConfiguration("None"))


    @staticmethod
    def _get_gcs_bucket_configuration(name: str, **kwargs) -> 'BucketConfiguration':
        return BucketConfiguration(name, BucketEncryptionConfiguration("None"))

EMPTY_BUCKET_CONFIGURATION = BucketConfiguration("", BucketEncryptionConfiguration("None"))