import logging
import time

import requests
from json.decoder import JSONDecodeError as json_exception
from requests import exceptions as requests_exceptions
from configparser import ConfigParser

from typing import Dict, Callable, Tuple, Union, Any

logger = logging.getLogger(__name__)

class HttpUtils:

    @staticmethod
    def do_api_call(config: ConfigParser, url: str, method: str, data: Dict[str, str] = None, params: Dict[str, str] = None, headers: Dict[str, str] = None, verify: bool = True, timeout: Union[int, Tuple[int, int]] = 10,
    max_tries: int = 10) -> Dict[str, str]:
        return HttpUtils._do_api_call_internal(HttpUtils._json, config, url, method, data, params, headers, verify, timeout=timeout, max_tries=max_tries)
    
    @staticmethod
    def maybe_api_call(config: ConfigParser, url: str, method: str, data: Dict[str, str] = None, params: Dict[str, str] = None, headers: Dict[str, str] = None, verify: bool = True, timeout: Union[int, Tuple[int, int]] = 10) -> Dict[str, str]:
        if config.getboolean('executor', 'skip_report', fallback=False) \
            or config.getboolean('general', 'unmanaged', fallback=False):
            logger.info("Skip report is set. Returning mock response")
            return {"response": "mock"}
        return HttpUtils._do_api_call_internal(HttpUtils._json, config, url, method, data, params, headers, verify, timeout=timeout)

    @staticmethod
    def do_plain_call(config: ConfigParser, url: str, method: str, data: Dict[str, str] = None, headers: Dict[str, str] = None, params: Dict[str, str] = None, verify: bool = True, timeout: Union[int, Tuple[int, int]] = 10) -> Dict[str, str]:
        return HttpUtils._do_api_call_internal(HttpUtils._plain, config, url, method, data, params, headers, verify, timeout=timeout)

    @staticmethod
    def _do_api_call_internal(operation: Callable, config: ConfigParser, url: str, method: str, data: Dict[str, str] = None, params: Dict[str, str] = None, headers: Dict[str, str] = None, verify: bool = True, timeout: Union[int, Tuple[int, int]] = 10,
    max_tries: int = 10) -> Dict[str, str]:
        attempts = 1
        while True:
            try:
                return operation(method, url, data, params, headers, verify, timeout=timeout)
            except requests_exceptions.RequestException as ex:
                time2sleep = 1
                if ex.response is not None:
                    status_code = ex.response.status_code
                    content = ex.response.reason
                else:
                    status_code = -1
                    content = "Unknown"
                if status_code == 403:
                    time2sleep = 20
                error_message = "Error making HTTP request. Code: {0} - Reason: {1} - Attempts: {2}".format(
                    status_code, content, attempts)
                if attempts > max_tries:
                    logger.error(error_message)
                    raise ex
                logger.warning("Error making HTTP request, retrying...")
                logger.warning(error_message)
                attempts += 1
                time.sleep(time2sleep)

    @staticmethod
    def _json(method: str, url: str, data: Dict[str, str], params: Dict[str, str], headers: Dict[str, str], verify: bool, timeout: Union[int, Tuple[int, int]] = 10) -> Dict[str, str]:
        response = requests.request(method, url, json=data, params=params, headers=headers, verify=verify, timeout=timeout)
        response.raise_for_status()
        try:
            return response.json()
        except json_exception:
            return {"response": response.text}

    @staticmethod
    def _plain(method: str, url: str, data: Dict[str, str], params: Dict[str, str], headers: Dict[str, str], verify: bool, timeout: Union[int, Tuple[int, int]] = 10) -> Dict[str, str]:
        response = requests.request(method, url, data=data, params=params, headers=headers, verify=verify, timeout=timeout)
        response.raise_for_status()
        r: Dict[str, Any] = {
            'headers': response.headers,
            'response': response.text
        }
        return r
        
