import json
import pathlib
import configparser
import logging

from typing import Any, Callable, Optional

from ec2init.plans.preinit.grow_data_volume import GrowDataVolume, NoDataVolumeException
from ec2init.plans.plan_utils import HopsPlanUtils

logger = logging.getLogger(__name__)

class BackupPropertyMissing(Exception):
    def __init__(self, property: str) -> None:
        super().__init__(f"Property {property} is missing from Backup Information")

class BackupImmutableProperty(Exception):
    def __init__(self, property: str) -> None:
        super().__init__(f"Property {property} is immutable")

class OffLineBackupInformation:

    EMPTY_STRING_CHECK: Callable[[str], bool] = lambda s : s != ""

    def __init__(self, config: Optional[configparser.ConfigParser]):
        self.config = config
        if config:
            try:
                # We only need it to detect if we are restoring a multi disk VM or not
                gv = GrowDataVolume(OffLineBackupInformation.__name__, config, HopsPlanUtils(config, None), True)
                gv._get_data_block_device(gv._get_data_mount_point())
                self.restoring_multidisk = True
                logger.info(f"{OffLineBackupInformation.__name__} - Restoring multi disk VM")
            except NoDataVolumeException:
                logger.info(f"{OffLineBackupInformation.__name__} - Restoring single disk VM")
                self.restoring_multidisk = False
        else:
            self.restoring_multidisk = False

    @property
    def ip(self) -> Optional[str]:
        if self.restoring_multidisk:
            return None
        self._has_property("_ip")
        return self._ip

    @ip.setter
    def ip(self, ip: str) -> None:
        self._is_property_set("_ip", OffLineBackupInformation.EMPTY_STRING_CHECK)
        self._ip = ip

    @property
    def hostname(self) -> Optional[str]:
        if self.restoring_multidisk:
            return None
        self._has_property("_hostname")
        return self._hostname

    @hostname.setter
    def hostname(self, hostname: str) -> None:
        self._is_property_set("_hostname", OffLineBackupInformation.EMPTY_STRING_CHECK)
        self._hostname = hostname

    @property
    def user_registry_url(self) -> Optional[str]:
        try:
            self._has_property("_user_registry_url")
        except BackupPropertyMissing:
            return None
        return self._user_registry_url

    @user_registry_url.setter
    def user_registry_url(self, user_registry_url: str) -> None:
        self._is_property_set("_user_registry_url", OffLineBackupInformation.EMPTY_STRING_CHECK)
        self._user_registry_url = user_registry_url

    def save(self, file_path: pathlib.Path) -> None:
        with open(file_path, 'w') as fd:
            json.dump(self, fd, cls=BackupInformationEncoder)

    def load(self, file_path: pathlib.Path) -> None:
        try:    
            bi = self._load_json_backup_info(file_path)
        except json.JSONDecodeError:
            bi = self._load_plaintext_backup_info(file_path)
        self.ip = bi.ip
        self.hostname = bi.hostname
        # optional properties
        try:
            self.user_registry_url = bi.user_registry_url
        except BackupPropertyMissing:
            pass

    def _is_property_set(self, property: str, check_func: Callable[[Any], bool]) -> None:
        if hasattr(self, property) and check_func(getattr(self, property)):
            raise BackupImmutableProperty(property[1:])

    def _has_property(self, property: str) -> None:
        if not hasattr(self, property):
            # remove _ from property name
            raise BackupPropertyMissing(property[1:])
        
    def _load_json_backup_info(self, file_path: pathlib.Path) -> 'OffLineBackupInformation':
        with open(file_path, 'r') as fd:
            return json.load(fd, object_hook=_decoder)

    def _load_plaintext_backup_info(self, file_path: pathlib.Path) -> 'OffLineBackupInformation':
        i = OffLineBackupInformation(None)
        with open(file_path, 'r') as fd:
            i.ip = self._strip_new_line(fd.readline())
            i.hostname = self._strip_new_line(fd.readline())
            user_registry_url = self._strip_new_line(fd.readline())
            if user_registry_url != "":
                i.user_registry_url = user_registry_url
        return i

    def _strip_new_line(self, text: str) -> str:
        return text.strip("\n")


    @staticmethod
    def is_backup(backupInfo: Optional['OffLineBackupInformation']) -> bool:
        if backupInfo:
            if hasattr(backupInfo, "_ip"):
                return backupInfo._ip is not None
        return False

def _decoder(o: Any) -> Any:
    i = OffLineBackupInformation(None)
    i.ip = o['_ip']
    i.hostname = o['_hostname']
    if '_user_registry_url' in o:
        i.user_registry_url = o['_user_registry_url']
    return i

class BackupInformationEncoder(json.JSONEncoder):
    def default(self, o: OffLineBackupInformation) -> Any:
        attrs = o.__dict__
        # we do not want to encode these attributes
        del attrs['config']
        del attrs['restoring_multidisk']
        return attrs